"""Utility functions for the Pythia protocol code."""

import logging
import threading
from typing import Any, Generic, Optional, TypeVar
from vizier import pythia
from vizier import pyvizier
from vizier._src.pyvizier.shared import common
from google.protobuf import any_pb2

_TT = TypeVar('_TT')


class ResponseWaiter(Generic[_TT]):
  """A stored message with a Wait() mechanism.

  This is a bridge between two threads; it stores a message, and
  the reader thread waits until the writer thread calls Report(..., done=True).
  Once Report() is called, it unblocks WaitForRespose().
  """

  def __init__(self):
    self._lock = threading.Lock()
    self._response: Optional[_TT] = None
    self._wait = threading.Event()

  def Report(self, update: _TT):
    """Called by the gRPC thread with a message from Vizier.

    When called, this unblocks WaitForResponse().

    Args:
      update:

    Raises:
      PythiaProtocolError
    """
    logging.info('About to take _lock in ResponseWaiter.Report()')
    with self._lock:
      if self._wait.is_set():
        raise pythia.PythiaProtocolError(
            'ResponseWaiter.Report() called after wait is set')
      self._response = update
      self._wait.set()

  def WaitForResponse(self) -> _TT:
    """Returns the result or raises an error.

    Raises:
      PythiaProtocolError: Due to a failure of the Pythia protocol.
      VizierDatabaseError: Vizier was unable to service the request.
    """
    self._wait.wait()
    logging.info('About to take _lock in ResponseWaiter.WaitForResponse()')
    with self._lock:
      if not self._response:
        logging.info('Raise')
        raise pythia.PythiaProtocolError('No response.')
      if self._response.error_details:
        raise pythia.VizierDatabaseError(self._response.error_details)
      logging.info('No raise -- WaitForResponse done')
      return self._response

  def CancelWait(self) -> None:
    """Cancel the wait."""
    logging.info('About to CancelWait()')
    self._wait.set()


def AssignKeyValuePlus(container: Any, *, trial_id: Optional[int], key: str,
                       ns: common.Namespace,
                       value: pyvizier.MetadataValue) -> None:
  """Insert or assign (key, value) to container.metadata.

  Args:
    container: A container class (e.g. a protobuf) that contains a $metadata
      field.  The $metadata field should be type KeyValuePlus.
    trial_id: To become container.metadata[i].trial_id; when unset, it means
      Study-wide metadata not associated with a Trial.
    key: To become container.metadata[i].key.
    ns: To become container.metadata[i].ns.
    value: To become either the $value or the $proto field of
      container.metadata[i], depending on the type of $value.   If $value is a
      string, it is inserted into ....value; else if it is a protobuf, it is
      packed into ....proto.
  """
  ns: str = ns.encode()
  str_trial_id: Optional[str] = str(trial_id) if trial_id is not None else None
  # The key is already in the metadata.
  for kv in container.metadata:
    trial_match = (not kv.HasField('trial_id')
                   if trial_id is None else kv.trial_id == str_trial_id)
    if trial_match and kv.k_v.key == key and kv.k_v.ns == ns:
      if isinstance(value, str):
        kv.ClearField('proto')
        kv.value = value
      elif isinstance(value, any_pb2.Any):
        kv.ClearField('value')
        kv.proto.CopyFrom(value)
      else:
        kv.ClearField('value')
        kv.proto.Pack(value)
      return

  # The key does not yet exist in the metadata.
  new_item = container.metadata.add(trial_id=str_trial_id)
  new_kv = new_item.k_v
  new_kv.key = key
  new_kv.ns = ns
  if isinstance(value, str):
    new_kv.value = value
  elif isinstance(value, any_pb2.Any):
    new_kv.proto = value
  else:
    new_kv.proto.Pack(value)



"""Contains all datastore classes for saving study + trial data.

See resources.py for naming conventions.
"""
import abc
import copy
import dataclasses
from typing import Callable, Dict, Iterable, List, Optional
from absl import logging

from vizier.service import key_value_pb2
from vizier.service import resources
from vizier.service import study_pb2
from vizier.service import vizier_oss_pb2
from vizier.service import vizier_service_pb2
from google.longrunning import operations_pb2

_KeyValuePlus = vizier_service_pb2.UpdateMetadataRequest.KeyValuePlus


class AlreadyExistsError(ValueError):
  """The resource key already exists in the database."""
  pass


class NotFoundError(KeyError):
  """The resource key does not exist in the database."""
  pass


class DataStore(abc.ABC):
  """Abstract class for data storage.

  Input/Outputs should always be pass-by-value.
  """

  @abc.abstractmethod
  def create_study(self, study: study_pb2.Study) -> resources.StudyResource:
    """Creates study in database. If preexisting, raises AlreadyExistsError."""

  @abc.abstractmethod
  def load_study(self, study_name: str) -> study_pb2.Study:
    """Loads study from the database. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def delete_study(self, study_name: str) -> None:
    """Deletes study from database. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def list_studies(self, owner_name: str) -> List[study_pb2.Study]:
    """Lists all studies under owner.


    Args:
      owner_name: Name of owner.

    Returns:
      List of Studies associated with owner.

    Raises:
      NotFoundError: If the owner does not exist.
    """

  @abc.abstractmethod
  def create_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    """Stores trial in database.

    Args:
      trial: Trial to be stored.

    Returns:
      The newly created trial resource name.

    Raises:
      AlreadyExistsError: If trial already exists.
    """

  @abc.abstractmethod
  def get_trial(self, trial_name: str) -> study_pb2.Trial:
    """Retrieves trial from database. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def update_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    """Updates pre-existing trial. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def list_trials(self, study_name: str) -> List[study_pb2.Trial]:
    """List all trials for study. If study nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def delete_trial(self, trial_name: str) -> None:
    """Deletes trial from database. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def max_trial_id(self, study_name: str) -> int:
    """Maximal trial ID in study (defaults to 0 if no trials exist).

    Args:
      study_name: Name of study.

    Returns:
      Maximum trial ID as an int.

    Raises:
      NotFoundError: If the study does not exist.
    """

  @abc.abstractmethod
  def create_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    """Stores suggestion operation.

    Args:
      operation:

    Returns:
      Resource for created op.

    Raises:
      AlreadyExistsError: If the suggest op already exists.
    """

  @abc.abstractmethod
  def get_suggestion_operation(self,
                               operation_name: str) -> operations_pb2.Operation:
    """Retrieves suggestion operation. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def update_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    """Updates pre-existing suggestion op.

    Args:
      operation: New suggestion speration.

    Returns:
      Resource to operation.

    Raises:
      NotFoundError if suggest op is nonexistent.
    """

  @abc.abstractmethod
  def list_suggestion_operations(
      self,
      owner_name: str,
      client_id: str,
      filter_fn: Optional[Callable[[operations_pb2.Operation], bool]] = None
  ) -> List[operations_pb2.Operation]:
    """Retrieve all suggestion op from client.

    Args:
      owner_name: Associated owner for the suggest op.
      client_id: Associated client for the suggest op.
      filter_fn: Optional function to filter out the suggest ops.

    Raises:
      NotFoundError: If owner or client is nonexistent.
    """

  @abc.abstractmethod
  def max_suggestion_operation_number(self, owner_name: str,
                                      client_id: str) -> int:
    """Maximal suggestion number for given client.

    Args:
      owner_name: Name of associated owners.
      client_id: ID of associated client.

    Returns:
      Maximum suggest op number as an int.

    Raises:
      NotFoundError: If owner or client is nonexistent.
    """

  @abc.abstractmethod
  def create_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    """Stores early stopping operation.

    Args:
      operation: Operation to be stored.

    Returns:
      The newly created op resource name.

    Raises:
      AlreadyExistsError: If the suggest op already exists.
    """

  @abc.abstractmethod
  def get_early_stopping_operation(
      self, operation_name: str) -> vizier_oss_pb2.EarlyStoppingOperation:
    """Retrieves early stopping op. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def update_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    """Updates preexisting early stopping op.

    Args:
      operation: New version of EarlyStoppingOp.

    Returns:
      Resource to the EarlyStoppingOp.

    Raises:
      NotFoundError: If early stopping op is nonexistent.
    """

  @abc.abstractmethod
  def update_metadata(
      self,
      study_name: str,
      study_metadata: Iterable[key_value_pb2.KeyValue],
      trial_metadata: Iterable[_KeyValuePlus],
  ) -> None:
    """Store the supplied metadata in the database.

    Args:
      study_name: (Typically derived from a StudyResource.)
      study_metadata: Metadata to attach to the Study as a whole.
      trial_metadata: Metadata to attach to Trials.

    Raises:
      NotFoundError: If the update fails because of an attempt to attach
        metadata to a nonexistant Trial.
    """


# Specific dataclasses used for NestedDictRAMDataStore.
@dataclasses.dataclass(frozen=True)
class StudyNode:
  """Contains original study and currently stored trials."""
  study_proto: study_pb2.Study

  # Keys are `trial_id`.
  trial_protos: Dict[int,
                     study_pb2.Trial] = dataclasses.field(default_factory=dict)

  # Keys are `operation_id`.
  early_stopping_operations: Dict[
      str, vizier_oss_pb2.EarlyStoppingOperation] = dataclasses.field(
          default_factory=dict)


@dataclasses.dataclass(frozen=True)
class ClientNode:
  """Only contains suggestion operations associated with this client."""
  # Keys are `operation_id`.
  suggestion_operations: Dict[str,
                              operations_pb2.Operation] = dataclasses.field(
                                  default_factory=dict)


@dataclasses.dataclass(frozen=True)
class OwnerNode:
  """First level of the entire RAM Datastore."""
  # Key is a `study_id` that pick's out one of the owner's Studies.
  studies: Dict[str, StudyNode] = dataclasses.field(default_factory=dict)
  # Key is `client_id`, which distinguishes clients, so you can have several
  # clients interacting with the same Vizier server.
  clients: Dict[str, ClientNode] = dataclasses.field(default_factory=dict)


class NestedDictRAMDataStore(DataStore):
  """Basic Datastore class using nested dictionaries."""

  def __init__(self):
    """Organization as follows."""
    # Key is `owner_id`, which corresponds to the (perhaps human) owner
    # of the Study.
    self._owners: Dict[str, OwnerNode] = {}

  def create_study(self, study: study_pb2.Study) -> resources.StudyResource:
    resource = resources.StudyResource.from_name(study.name)
    temp_dict = {resource.study_id: StudyNode(study_proto=copy.deepcopy(study))}

    if resource.owner_id not in self._owners:
      self._owners[resource.owner_id] = OwnerNode(studies=temp_dict)
    else:
      study_dict = self._owners[resource.owner_id].studies
      if resource.study_id not in study_dict:
        study_dict.update(temp_dict)
      else:
        raise AlreadyExistsError('Study with that name already exists.',
                                 study.name)
    return resource

  def load_study(self, study_name: str) -> study_pb2.Study:
    resource = resources.StudyResource.from_name(study_name)
    try:
      return copy.deepcopy(self._owners[resource.owner_id].studies[
          resource.study_id].study_proto)
    except KeyError as err:
      raise NotFoundError('Could not get Study with name:',
                          resource.name) from err

  def delete_study(self, study_name: str) -> None:
    resource = resources.StudyResource.from_name(study_name)
    try:
      del self._owners[resource.owner_id].studies[resource.study_id]
    except KeyError as err:
      raise NotFoundError('Study does not exist:', study_name) from err

  def list_studies(self, owner_name: str) -> List[study_pb2.Study]:
    resource = resources.OwnerResource.from_name(owner_name)
    try:
      study_nodes = list(self._owners[resource.owner_id].studies.values())
      return copy.deepcopy(
          [study_node.study_proto for study_node in study_nodes])
    except KeyError as err:
      raise NotFoundError('Owner does not exist:', owner_name) from err

  def create_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    resource = resources.TrialResource.from_name(trial.name)
    trial_protos = self._owners[resource.owner_id].studies[
        resource.study_id].trial_protos
    if resource.trial_id in trial_protos:
      raise AlreadyExistsError('Trial %s already exists' % trial.name)
    else:
      trial_protos[resource.trial_id] = copy.deepcopy(trial)
    return resource

  def get_trial(self, trial_name: str) -> study_pb2.Trial:
    resource = resources.TrialResource.from_name(trial_name)
    try:
      return copy.deepcopy(self._owners[resource.owner_id].studies[
          resource.study_id].trial_protos[resource.trial_id])
    except KeyError as err:
      raise NotFoundError('Could not get Trial with name:',
                          resource.name) from err

  def update_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    resource = resources.TrialResource.from_name(trial.name)
    try:
      trial_protos = self._owners[resource.owner_id].studies[
          resource.study_id].trial_protos
      if resource.trial_id not in trial_protos:
        raise NotFoundError('Trial %s does not exist.' % trial.name)
      trial_protos[resource.trial_id] = copy.deepcopy(trial)
      return resource
    except KeyError as err:
      raise NotFoundError('Could not update Trial with name:',
                          resource.name) from err

  def list_trials(self, study_name: str) -> List[study_pb2.Trial]:
    resource = resources.StudyResource.from_name(study_name)
    try:
      return copy.deepcopy(
          list(self._owners[resource.owner_id].studies[
              resource.study_id].trial_protos.values()))
    except KeyError as err:
      raise NotFoundError('Study does not exist:', study_name) from err

  def delete_trial(self, trial_name: str) -> None:
    resource = resources.TrialResource.from_name(trial_name)
    try:
      del self._owners[resource.owner_id].studies[
          resource.study_id].trial_protos[resource.trial_id]
    except KeyError as err:
      raise NotFoundError('Trial does not exist:', trial_name) from err

  def max_trial_id(self, study_name: str) -> int:
    resource = resources.StudyResource.from_name(study_name)
    try:
      trial_ids = list(self._owners[resource.owner_id].studies[
          resource.study_id].trial_protos.keys())
    except KeyError as err:
      raise NotFoundError('Study does not exist:', study_name) from err

    if trial_ids:
      return max(trial_ids)
    else:
      return 0

  def create_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)

    if resource.client_id not in self._owners[resource.owner_id].clients:
      self._owners[resource.owner_id].clients[resource.client_id] = ClientNode()
    suggestion_operations = self._owners[resource.owner_id].clients[
        resource.client_id].suggestion_operations

    if resource.operation_id in suggestion_operations:
      raise AlreadyExistsError('Operation already exists:',
                               resource.operation_id)

    suggestion_operations[resource.operation_id] = copy.deepcopy(operation)
    return resource

  def get_suggestion_operation(self,
                               operation_name: str) -> operations_pb2.Operation:
    resource = resources.SuggestionOperationResource.from_name(operation_name)
    try:
      return copy.deepcopy(self._owners[resource.owner_id].clients[
          resource.client_id].suggestion_operations[resource.operation_id])

    except KeyError as err:
      raise NotFoundError('Could not find SuggestionOperation with name:',
                          resource.name) from err

  def update_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)
    try:
      self._owners[resource.owner_id].clients[
          resource.client_id].suggestion_operations[
              resource.operation_id] = copy.deepcopy(operation)
      return resource
    except KeyError as err:
      raise NotFoundError('Could not update SuggestionOperation with name:',
                          resource.name) from err

  def list_suggestion_operations(
      self,
      owner_name: str,
      client_id: str,
      filter_fn: Optional[Callable[[operations_pb2.Operation], bool]] = None
  ) -> List[operations_pb2.Operation]:
    resource = resources.OwnerResource.from_name(owner_name)
    try:
      operations_list = list(self._owners[
          resource.owner_id].clients[client_id].suggestion_operations.values())
    except KeyError as err:
      raise NotFoundError('(owner_name, client_id) does not exist:',
                          (owner_name, client_id)) from err

    if filter_fn is not None:
      return copy.deepcopy([op for op in operations_list if filter_fn(op)])
    else:
      return copy.deepcopy(operations_list)

  def max_suggestion_operation_number(self, owner_name: str,
                                      client_id: str) -> int:
    resource = resources.OwnerResource.from_name(owner_name)
    try:
      ops = self._owners[
          resource.owner_id].clients[client_id].suggestion_operations
    except KeyError as err:
      raise NotFoundError('(owner_name, client_id) does not exist:',
                          (owner_name, client_id)) from err
    return len(ops)

  def create_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name)

    early_stopping_ops = self._owners[resource.owner_id].studies[
        resource.study_id].early_stopping_operations
    if resource.operation_id in early_stopping_ops:
      raise AlreadyExistsError('Operation already exists:',
                               resource.operation_id)

    early_stopping_ops[resource.operation_id] = copy.deepcopy(operation)
    return resource

  def get_early_stopping_operation(
      self, operation_name: str) -> vizier_oss_pb2.EarlyStoppingOperation:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation_name)
    try:
      return copy.deepcopy(self._owners[resource.owner_id].studies[
          resource.study_id].early_stopping_operations[resource.operation_id])
    except KeyError as err:
      raise NotFoundError('Could not find EarlyStoppingOperation with name:',
                          resource.name) from err

  def update_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name)
    try:
      self._owners[resource.owner_id].studies[
          resource.study_id].early_stopping_operations[
              resource.operation_id] = copy.deepcopy(operation)
      return resource
    except KeyError as err:
      raise NotFoundError('Could not update EarlyStoppingOperation with name:',
                          resource.name) from err

  def update_metadata(self, study_name: str,
                      study_metadata: Iterable[key_value_pb2.KeyValue],
                      trial_metadata: Iterable[_KeyValuePlus]) -> None:
    """Writes the supplied metadata to the database.

    Args:
      study_name:
      study_metadata: Metadata that's associated with the Study as a whole.
      trial_metadata: Metadata that's associated with trials.  (Note that its an
        error to attach metadata to a Trial that doesn't exist.)
    """
    s_resource = resources.StudyResource.from_name(study_name)
    logging.debug('database.update_metadata s_resource= %s', s_resource)

    try:
      study_node = self._owners[s_resource.owner_id].studies[
          s_resource.study_id]
    except KeyError as e:
      raise NotFoundError('No such study:', s_resource.name) from e
    # Store Study-related metadata into the database.
    study_node.study_proto.study_spec.ClearField('metadata')
    for metadata in study_metadata:
      study_node.study_proto.study_spec.metadata.append(metadata)
    # Store trial-related metadata in the database. We first create a dict of
    # the relevant `trial_resources` that will be touched. We clear them, then
    # loop through the metadata, converting to protos.
    trial_resources: Dict[str, resources.TrialResource] = {}
    for metadata in trial_metadata:
      if metadata.trial_id in trial_resources:
        t_resource = trial_resources[metadata.trial_id]
      else:
        # If we don't have a t_resource entry already, create one and clear the
        # relevant Trial's metadata.
        t_resource = s_resource.trial_resource(metadata.trial_id)
        trial_resources[metadata.trial_id] = t_resource
        try:
          study_node.trial_protos[t_resource.trial_id].ClearField('metadata')
        except KeyError as e:
          raise NotFoundError(f'No such trial ({metadata.trial_id}):',
                              t_resource.name) from e
      study_node.trial_protos[t_resource.trial_id].metadata.append(metadata.k_v)



"""Tests for sql_datastore."""
import sqlalchemy as sqla

from vizier.service import datastore_test_lib
from vizier.service import sql_datastore
from vizier.service import test_util
from absl.testing import absltest


class SQLDataStoreTest(datastore_test_lib.DataStoreTestCase):

  def setUp(self):
    self.owner_id = 'my_username'
    self.study_id = '123123123'
    self.client_id = 'client_0'
    self.example_study = test_util.generate_study(self.owner_id, self.study_id)
    self.example_trials = test_util.generate_trials([1, 2],
                                                    owner_id=self.owner_id,
                                                    study_id=self.study_id)
    self.example_suggestion_operations = test_util.generate_suggestion_operations(
        [1, 2, 3, 4], self.owner_id, self.client_id)
    self.example_early_stopping_operations = test_util.generate_early_stopping_operations(
        [1, 2], self.owner_id, self.study_id)

    engine = sqla.create_engine('sqlite:///:memory:', echo=True)
    self.datastore = sql_datastore.SQLDataStore(engine)
    super().setUp()

  def test_study_api(self):
    self.assertStudyAPI(self.datastore, self.example_study)

  def test_trial(self):
    self.assertTrialAPI(self.datastore, self.example_study, self.example_trials)

  def test_suggestion_operation(self):
    self.assertSuggestOpAPI(self.datastore, self.example_study, self.client_id,
                            self.example_suggestion_operations)

  def test_early_stopping_operation(self):
    self.assertEarlyStoppingAPI(self.datastore, self.example_study,
                                self.example_trials,
                                self.example_early_stopping_operations)

  def test_update_metadata(self):
    self.assertUpdateMetadataAPI(self.datastore, self.example_study,
                                 self.example_trials)


if __name__ == '__main__':
  absltest.main()



"""Tests for designer_policy."""

import copy

from typing import Optional, Sequence
from vizier import algorithms as vza
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.policies import designer_policy as dp
from vizier.interfaces import serializable
from absl.testing import absltest


class _FakeSerializableDesigner(vza.PartiallySerializableDesigner,
                                vza.SerializableDesigner):
  # If set, load() and recover() raise the provided Exception.
  _error_on_load: Optional[Exception] = None

  def __init__(self):
    self._num_incorporated_trials = 0

    # For debugging and testing. Not stored in the metadata.
    self._last_delta = None

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    return [vz.TrialSuggestion(vz.ParameterDict())] * count

  def update(self, delta: vza.CompletedTrials):
    self._last_delta = delta
    self._num_incorporated_trials += len(delta.completed)

  def dump(self) -> vz.Metadata:
    md = vz.Metadata()
    md.ns('ns1')['foo1'] = 'bar1'
    md.ns('ns1').ns('ns11')['foo11'] = 'bar11'
    md['num_incorporated_trials'] = str(self._num_incorporated_trials)
    return md

  @classmethod
  def recover(cls, md: vz.Metadata) -> '_FakeSerializableDesigner':
    if cls._error_on_load is not None:
      raise cls._error_on_load  # pylint:disable=raising-bad-type
    try:
      designer = cls()
      designer._num_incorporated_trials = int(md['num_incorporated_trials'])
      return designer
    except KeyError as e:
      raise KeyError(f'Cannot find in {md}') from e

  def load(self, md: vz.Metadata):
    if self._error_on_load is not None:
      raise self._error_on_load  # pylint:disable=raising-bad-type
    try:
      self._num_incorporated_trials = int(md['num_incorporated_trials'])
    except KeyError as e:
      raise serializable.DecodeError(f'Cannot find in {md}') from e


# TODO: Add more tests

_NUM_INITIAL_COMPLETED_TRIALS = 10


class DesignerPolicyNormalOperationTest(absltest.TestCase):
  """Tests Designer policies under error-free conditions."""

  def setUp(self):
    super().setUp()
    self.maxDiff = None  # pylint: disable=invalid-name
    designer = _FakeSerializableDesigner()
    runner = pythia.LocalPolicyRunner(vz.StudyConfig())
    runner.AddTrials([
        vz.Trial().complete(vz.Measurement())
        for _ in range(_NUM_INITIAL_COMPLETED_TRIALS)
    ])

    # Run with a policy
    policy = dp.PartiallySerializableDesignerPolicy(
        runner, lambda _: designer, ns_root='test', verbose=2)
    trials = runner.SuggestTrials(policy, 5)
    self.assertLen(designer._last_delta.completed,
                   _NUM_INITIAL_COMPLETED_TRIALS)
    self.assertEqual(
        runner.study_descriptor().config.metadata.ns('test').ns('designer')
        ['num_incorporated_trials'], str(_NUM_INITIAL_COMPLETED_TRIALS))

    # Complete trials
    for t in trials[::2]:
      t.complete(vz.Measurement())

    self.runner = runner
    self.designer = designer
    self.trials = trials

  def test_partially_serializable(self):
    runner, designer, trials = self.runner, self.designer, self.trials

    # Mimick the server environment by creating a new policy.
    policy = dp.PartiallySerializableDesignerPolicy(
        runner, lambda _: designer, ns_root='test', verbose=2)
    runner.SuggestTrials(policy, 1)
    # Delta should consist only of the newly completed trials.
    self.assertSequenceEqual(designer._last_delta.completed, trials[::2])
    self.assertEqual(designer._num_incorporated_trials, 13)
    self.assertEqual(
        runner.study_descriptor().config.metadata.ns('test').ns('designer')
        ['num_incorporated_trials'], str(13))

  def test_fully_serializable(self):
    runner, _, trials = self.runner, self.designer, self.trials

    # Mimick the server environment by creating a new policy.
    def raise_error(*args, **kwargs):
      raise ValueError('This code should not be called')

    policy = dp.SerializableDesignerPolicy(
        runner,
        designer_factory=raise_error,  # should not be used.
        designer_cls=_FakeSerializableDesigner,
        ns_root='test',
        verbose=2)
    runner.SuggestTrials(policy, 1)

    # `policy` creates a new designer object by loading from metadata.
    # Delta should consist only of the newly completed trials.
    new_designer = policy.designer
    self.assertSequenceEqual(new_designer._last_delta.completed, trials[::2])
    self.assertEqual(new_designer._num_incorporated_trials, 13)
    self.assertEqual(
        runner.study_descriptor().config.metadata.ns('test').ns('designer')
        ['num_incorporated_trials'], str(13))

  def test_non_serializable(self):
    runner, designer, trials = self.runner, self.designer, self.trials

    # Mimick the server environment by creating a new policy.
    policy = dp.DesignerPolicy(copy.deepcopy(runner), lambda _: designer)
    runner.SuggestTrials(policy, 1)
    # Designer is updated with all trials, including both the initial batch
    # of completed trials and the newly completed trials.
    self.assertLen(designer._last_delta.completed,
                   _NUM_INITIAL_COMPLETED_TRIALS + len(trials[::2]))


if __name__ == '__main__':
  absltest.main()



"""Utility functions for handling vizier metadata."""

from typing import Tuple, Union, Optional, TypeVar, Type
from vizier.service import key_value_pb2
from vizier.service import study_pb2
from google.protobuf import any_pb2
from google.protobuf.message import Message

T = TypeVar('T')


def assign(
    container: Union[study_pb2.StudySpec, study_pb2.Trial], *, key: str,
    ns: str, value: Union[str, any_pb2.Any,
                          Message]) -> Tuple[key_value_pb2.KeyValue, bool]:
  """Insert or assign (key, value) to container.metadata.

  Args:
    container: container.metadata must be repeated KeyValue (protobuf) field.
    key:
    ns: A namespace for the key (defaults to '', which is the user's namespace).
    value: Behavior depends on the type. `str` is copied to KeyValue.value
      `any_pb2.Any` is copied to KeyValue.proto Other types are packed to
      any_pb2.Any proto, which is then copied to KeyValue.proto.

  Returns:
    (proto, inserted) where
    proto is the protobuf that was just inserted into the $container, and
    inserted is True if the proto was newly inserted, False if it was replaced.
  """

  for kv in container.metadata:
    if kv.key == key and kv.ns == ns:
      if isinstance(value, str):
        kv.ClearField('proto')
        kv.value = value
      elif isinstance(value, any_pb2.Any):
        kv.ClearField('value')
        kv.proto.CopyFrom(value)
      else:
        kv.ClearField('value')
        kv.proto.Pack(value)
      return kv, False

  # The key does not exist in the metadata.
  if isinstance(value, str):
    metadata = container.metadata.add(key=key, ns=ns, value=value)
  elif isinstance(value, any_pb2.Any):
    metadata = container.metadata.add(key=key, ns=ns, proto=value)
  else:
    metadata = container.metadata.add(key=key, ns=ns)
    metadata.proto.Pack(value)
  return metadata, True


def get(container: Union[study_pb2.StudySpec, study_pb2.Trial], *, key: str,
        ns: str) -> Optional[str]:
  """Returns the metadata value associated with key, or None.

  Args:
    container: A Trial of a StudySpec in protobuf form.
    key: The key of a KeyValue protobuf.
    ns: A namespace for the key (defaults to '', which is the user's namespace).
  """

  for kv in container.metadata:
    if kv.key == key and kv.ns == ns:
      if not kv.HasField('proto'):
        return kv.value
  return None


def get_proto(container: Union[study_pb2.StudySpec, study_pb2.Trial], *,
              key: str, ns: str, cls: Type[T]) -> Optional[T]:
  """Unpacks the proto metadata into message.

  Args:
    container: (const) StudySpec or Trial to search the metadata from.
    key: (const) Lookup key of the metadata.
    ns: A namespace for the key (defaults to '', which is the user's namespace).
    cls: Pass in a proto ***class***, not a proto object.

  Returns:
    Proto message, if the value associated with the key exists and
    can be parsed into proto; None otherwise.
  """
  for kv in container.metadata:
    if kv.key == key and kv.ns == ns:
      if kv.HasField('proto'):
        message = cls()
        success = kv.proto.Unpack(message)
        return message if success else None
  return None



"""Wraps Designer into Policy."""
import abc
import json
from typing import Callable, Generic, Sequence, Type, TypeVar

from absl import logging
from vizier import algorithms as vza
from vizier import pythia
from vizier import pyvizier as vz
from vizier.interfaces import serializable

_T = TypeVar('_T')
Factory = Callable[[vz.StudyConfig], _T]


class DesignerPolicy(pythia.Policy):
  """Wraps a Designer into pythia Policy."""

  def __init__(self, supporter: pythia.PolicySupporter,
               designer_factory: Factory[vza.Designer]):
    self._supporter = supporter
    self._designer_factory = designer_factory

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecisions:
    self._designer = self._designer_factory(request.study_config)
    new_trials = self._supporter.GetTrials(
        status_matches=vz.TrialStatus.COMPLETED)
    self._designer.update(vza.CompletedTrials(new_trials))

    return pythia.SuggestDecisions.from_trials(
        self._designer.suggest(request.count))


class _SerializableDesignerPolicyBase(pythia.Policy,
                                      serializable.PartiallySerializable,
                                      Generic[_T], abc.ABC):
  """Wraps a PartiallySerializable into pythia Policy."""

  _ns_designer = 'designer'

  def __init__(self,
               supporter: pythia.PolicySupporter,
               designer_factory: Callable[[vz.StudyConfig], _T],
               *,
               ns_root: str = 'designer_policy_v0',
               verbose: int = 0):
    """Init.

    Args:
      supporter:
      designer_factory:
      ns_root: Root of the namespace where policy state is stored.
      verbose: Logging verbosity.
    """
    self._supporter = supporter
    self._designer_factory = designer_factory
    self._ns_root = ns_root
    self._incorporated_trial_ids = set()
    self._reference_study_config = supporter.GetStudyConfig()
    self._verbose = verbose
    self._designer = None

  @property
  def designer(self) -> _T:
    if self._designer is None:
      raise ValueError('`self._designer` has not been initialized!'
                       'Use self._create_designer(..) to initialize it.')
    return self._designer

  @abc.abstractmethod
  def _create_designer(self, designer_metadata: vz.Metadata) -> _T:
    """Creates a new Designer by restoring the state from `designer_metadata`.

    Args:
      designer_metadata:

    Returns:
      New Designer object.

    Raises:
      DecodeError: `designer_metadata` does not contain valid information
        to restore a Designer state.
    """
    pass

  # TODO: Use timestamps to avoid metadata blowup.
  def load(self, md: vz.Metadata) -> None:
    if 'incorporated_trial_ids' in md:
      try:
        self._incorporated_trial_ids = set(
            json.loads(md['incorporated_trial_ids']))
      except json.JSONDecodeError as e:
        raise serializable.HarmlessDecodeError from e
    else:
      raise serializable.HarmlessDecodeError()
    self._log(
        self._verbose,
        'Successfully recovered the policy state, which incorporated %s trials',
        len(self._incorporated_trial_ids))
    self._designer = self._create_designer(md.ns(self._ns_designer))

  def _log(self, level, *args):
    if level <= self._verbose:
      logging.info(*args)

  def _initialize_designer(self, study_config: vz.StudyConfig) -> None:
    """Populates `self._designer` with a new designer.

    This method catches all DecodeErrors and guarantees that `self.designer`
    does not raise an Exception.

    Args:
      study_config:

    Raises:
      ValueError: If study_config is differerent from the initially
        receieved study_config.
    """
    if self._designer is not None:
      # When the same policy object is maintained in RAM, prefer keeping
      # the designer object over restoring the state from metadata.
      # TOCONSIDER: Adding a boolean knob to turn off this behavior.
      self._log(
          2,
          "Policy already has a designer, and won't attempt to load from metadata"
      )
      return
    elif self._reference_study_config != study_config:
      raise ValueError(
          f'{type(self)} cannot be re-used for different study configs!'
          f'Policy: {self}, previous study: {self._reference_study_config} '
          f'new study: {study_config}')

    metadata = study_config.metadata.ns(self._ns_root)
    try:
      self.load(metadata)
      self._log(1, 'Succesfully decoded all states!',
                len(self._incorporated_trial_ids))
    except serializable.DecodeError as e:
      self._log(1, 'Failed to decode state. %s', e)
      self._designer = self._designer_factory(study_config)
      self._incorporated_trial_ids = set()

  def dump(self) -> vz.Metadata:
    """Dump state.

    Returns:
      Metadata has the following namespace hierarchy:
        Namespace([self._ns_root]): contains the policy's state.
        Namespace([self._ns_root, self._ns_designer]: contains the designer's
          state.
    """

    md = vz.Metadata()
    md.ns(self._ns_designer).attach(self.designer.dump())
    # TODO: Storing every id is inefficient. Optimize this.
    md['incorporated_trial_ids'] = json.dumps(
        list(self._incorporated_trial_ids))
    return md

  def _get_new_trials(self, max_trial_id: int) -> Sequence[vz.CompletedTrial]:
    """Returns new completed trials that designer should be updated with."""
    if len(self._incorporated_trial_ids) == max_trial_id:
      # no trials need to be loaded.
      return []
    all_trial_ids = set(range(1, max_trial_id + 1))
    trial_ids_to_load = all_trial_ids - self._incorporated_trial_ids

    trials = self._supporter.GetTrials(
        trial_ids=trial_ids_to_load, status_matches=vz.TrialStatus.COMPLETED)
    self._log(
        1, 'Loaded %s completed trials out of %s total unseen trials. '
        'Max trial id is %s.', len(trials), len(trial_ids_to_load),
        max_trial_id)
    return trials

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecisions:
    # Note that we can avoid O(Num trials) dependency in the standard scenario,
    # by storing only the last element in a consecutive sequence, e.g.,
    # instead of storing [1,2,3,4,11,12,13,21], store: [4,13,21], but
    # we keep things simple in this pseudocode.
    self._initialize_designer(request.study_descriptor.config)
    new_trials = self._get_new_trials(request.study_descriptor.max_trial_id)
    self.designer.update(vza.CompletedTrials(new_trials))
    self._incorporated_trial_ids |= set(t.id for t in new_trials)

    self._log(
        1, 'Updated with %s trials. Designer has seen a total of %s trials.',
        len(new_trials), len(self._incorporated_trial_ids))

    with self._supporter.MetadataUpdate() as mu:
      # pylint: disable=protected-access
      # TODO: Improve the MetadataUpdateContext API.
      mu._delta.on_study.ns(self._ns_root).attach(self.dump())

    return pythia.SuggestDecisions.from_trials(
        self.designer.suggest(request.count))


class PartiallySerializableDesignerPolicy(
    _SerializableDesignerPolicyBase[vza.PartiallySerializableDesigner]):
  """Wraps a PartiallySerializableDesigner."""

  def _create_designer(
      self,
      designer_metadata: vz.Metadata) -> vza.PartiallySerializableDesigner:
    designer = self._designer_factory(self._reference_study_config)
    designer.load(designer_metadata)
    return designer


class SerializableDesignerPolicy(
    _SerializableDesignerPolicyBase[vza.SerializableDesigner]):
  """Wraps a SerializableDesigner."""

  def __init__(self,
               supporter: pythia.PolicySupporter,
               designer_factory: Factory[vza.SerializableDesigner],
               designer_cls: Type[vza.SerializableDesigner],
               *,
               ns_root: str = 'designer_policy_v0',
               verbose: int = 0):
    """Init.

    Args:
      supporter:
      designer_factory: Used when designer state cannot be restored.
      designer_cls: Type name of the designer. Its load() classmethod is called
        to restore the designer state.
      ns_root: Root of the namespace where policy state is stored.
      verbose: Logging verbosity.
    """
    super().__init__(
        supporter, designer_factory, ns_root=ns_root, verbose=verbose)
    self._designer_cls = designer_cls

  def _create_designer(
      self, designer_metadata: vz.Metadata) -> vza.SerializableDesigner:
    return self._designer_cls.recover(designer_metadata)



"""Exceptions used in the PythiaServer and Pythia Policy code."""


class PythiaError(Exception):
  """Abstract superclass of Pythia-related errors."""


class TemporaryPythiaError(PythiaError):
  """Vizier should attempt this interaction with a different Pythia task.

  Policy code should raise this to terminate and retry its
  interaction with Vizier.  Maps to an ABORTED gRPC error code.

  Use case: Catch-all errors where there's some hope that a re-run will succeed
    (e.g. OOM, timeouts, snapshot/version problems, or failure of
    nondeterministic algorithms).  If a Policy raises this error, Vizier will
    re-try the computation a modest number of times.
  """


class CachedPolicyIsStaleError(PythiaError):
  """Raise this if the cached Policy instance is too stale.

  Raised by Policy code.  Because Policy instances are cached,
  they could potentially become stale.  If so, code should raise this,
  and the computation will be restarted with a freshly created Policy
  instance.
  """


class InactivateStudyError(PythiaError):
  """Pythia cannot handle this Study as configured; inactivate it.

  This terminates the interaction and returns an ARGUMENT_INVALID gRPC error
  code.  It's intended for unrecoverable errors, or misconfigued Studies.
  There will be no re-tries.
  """


class LoadTooLargeError(PythiaError):
  """Raised when the Pythia server is overly busy.

  If a Policy raises this error, it will be retried a (nearly) infinite number
  of times by the Vizier server, on the assumption that there will eventually
  be a server that's not overly busy.
  """


class CancelledByVizierError(PythiaError):
  """Indicates that Vizier cancelled the interaction."""
  pass


class CancelComputeError(Exception):
  """Raised by the Pythia server to kill Policy computations.

  This is raised in the Pythia compute thread (it is raised by the
  PolicySupporter object) when Vizier cancels the interaction. Policy code
  may catch this temporarily to do clean-up actions, but should re-raise it.

  Vizier may cancel the interaction when the SuggestionWorker needs to shut down
  (i.e. because it's preempted), or when the Pythia computation runs too long.
  """


class PythiaProtocolError(Exception):
  """There was a problem with the interaction between Vizier and Pythia.

  This implies a bug in the protocol code.  Do not raise or catch this in
  Policy code; please report any occurrences to vizier-team@google.com.
  Maps to an INTERNAL gRPC error code.
  """


class VizierDatabaseError(Exception):
  """Vizier had an error on a database operation.

  This can mean an attempt to access unavailable data, access control, etc.
  """



"""Experimenter function implementations for BBOB functions."""

import functools
import hashlib
import math
from typing import Any, Callable

import numpy as np
from vizier import pyvizier


def DefaultBBOBProblemStatement(
    dimension: int,
    *,
    metric_name="bbob_eval",
    min_value: float = -5.0,
    max_value: float = 5.0) -> pyvizier.ProblemStatement:
  """Returns default BBOB ProblemStatement for given dimension."""
  problem_statement = pyvizier.ProblemStatement()
  space = problem_statement.search_space
  for dim in range(dimension):
    space.select_root().add_float_param(
        name=f"x{dim}", min_value=min_value, max_value=max_value)
  problem_statement.metric_information.append(
      pyvizier.MetricInformation(
          name=metric_name, goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
  return problem_statement


## Utility Functions for BBOB.
def LambdaAlpha(alpha: float, dim: int) -> np.ndarray:
  """The BBOB LambdaAlpha matrix creation function.

  Args:
    alpha: Function parameter.
    dim: Dimension of matrix created.

  Returns:
    Diagonal matrix of dimension dim with values determined by alpha.
  """
  lambda_alpha = np.zeros([dim, dim])
  for i in range(dim):
    exp = (0.5 * (float(i) / (dim - 1))) if dim > 1 else 0.5
    lambda_alpha[i, i] = alpha**exp
  return lambda_alpha


def ArrayMap(vector: np.ndarray, fn: Callable[[float], float]) -> np.ndarray:
  """Create a new array by mapping fn() to each element of the original array.

  Args:
    vector: ndarray to be mapped.
    fn: scalar function for mapping.

  Returns:
    New ndarray be values mapped by fn.
  """
  results = np.zeros(vector.shape)
  for i, v in enumerate(vector.flat):
    results.flat[i] = fn(v)
  return results


def Tosz(element: float) -> float:
  """The BBOB T_osz function.

  Args:
    element: float input.

  Returns:
    Tosz(input).
  """
  x_carat = 0.0 if element == 0 else math.log(abs(element))
  c1 = 10.0 if element > 0 else 5.5
  c2 = 7.9 if element > 0 else 3.1
  return np.sign(element) * math.exp(
      x_carat + 0.049 * (math.sin(c1 * x_carat) + math.sin(c2 * x_carat)))


def Tasy(vector: np.ndarray, beta: float) -> np.ndarray:
  """The BBOB Tasy function.

  Args:
    vector: ndarray
    beta: Function parameter

  Returns:
    ndarray with values determined by beta.
  """
  dim = len(vector)
  result = np.zeros([dim, 1])
  for i, val in enumerate(vector.flat):
    if val > 0:
      t = i / (dim - 1.0) if dim > 1 else 1
      exp = 1 + beta * t * (val**0.5)
    else:
      exp = 1
    result[i] = val**exp
  return result


def SIndex(dim: int, to_sz) -> float:
  """Calculate the BBOB s_i.

  Assumes i is 0-index based.

  Args:
    dim: dimension
    to_sz: values

  Returns:
    float representing SIndex(i, d, to_sz).
  """
  s = np.zeros([dim,])
  for i in range(dim):
    if dim > 1:
      s[i] = 10**(0.5 * (i / (dim - 1.0)))
    else:
      s[i] = 10**0.5
    if i % 2 == 0 and to_sz[i] > 0:
      s[i] *= 10
  return s


def Fpen(vector: np.ndarray) -> float:
  """The BBOB Fpen function.

  Args:
    vector: ndarray.

  Returns:
    float representing Fpen(vector).
  """
  return sum([max(0.0, (abs(x) - 5.0))**2 for x in vector.flat])


def _Hash(*seeds: Any) -> int:
  """Returns a stable hash that fits in a positive int64."""
  message = hashlib.sha256()
  for s in seeds:
    message.update(bytes(s))
  # Limit the size of the returned value to fit into a np.int64.
  return int(message.hexdigest()[:15], 16)


def _ToFloat(a: int, b: np.ndarray) -> np.ndarray:
  """Convert a%b where b is an int into a float on [-0.5, 0.5]."""
  return (np.int64(a) % b) / np.float64(b) - 0.5


@functools.lru_cache(maxsize=128)
def _R(dim: int, seed: int, *moreseeds: bytes) -> np.ndarray:
  """Returns an orthonormal rotation matrix.

  Args:
    dim: size of the resulting matrix.
    seed: int seed
    *moreseeds: Additional parameters to include in the hash.  Arguments must be
      interpretable as a buffer of bytes.

  Returns:
    2-dimensional, dim x dim, ndarray, representing a rotation matrix.
  """
  if seed == 0:
    return np.identity(dim)
  h = 0 if seed == 0 else _Hash(*((seed, dim) + moreseeds))
  a = np.arange(dim * dim, dtype=np.int64)
  # We make a vector of (loosely speaking) random entries.
  b = (_ToFloat(h + 17, a + 127) + _ToFloat(h + 61031, a + 197) +
       _ToFloat(h + 503, a + 293))

  # The "Q" part of a Q-R decomposition is orthonormal, so $result is a pure
  # rotation matrix.  If elements of $b are independent normal, then the
  # distribution of rotations is uniform over the hypersphere.
  return np.linalg.qr(b.reshape(dim, dim))[0]


## BBOB Functions.
def Sphere(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Sphere function."""
  del seed
  return float(np.sum(arr * arr))


def Rastrigin(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Rastrigin function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  z = np.matmul(_R(dim, seed, b"R"), arr)
  z = Tasy(ArrayMap(z, Tosz), 0.2)
  z = np.matmul(_R(dim, seed, b"Q"), z)
  z = np.matmul(LambdaAlpha(10.0, dim), z)
  z = np.matmul(_R(dim, seed, b"R"), z)
  return float(10 * (dim - np.sum(np.cos(2 * math.pi * z))) +
               np.sum(z * z, axis=0))


def BuecheRastrigin(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB BuecheRastrigin function."""
  del seed
  dim = len(arr)
  arr.shape = (dim, 1)
  t = ArrayMap(arr, Tosz)
  l = SIndex(dim, arr) * t.flat

  term1 = 10 * (dim - np.sum(np.cos(2 * math.pi * l), axis=0))
  term2 = np.sum(l * l, axis=0)
  term3 = 100 * Fpen(arr)
  return float(term1 + term2 + term3)


def LinearSlope(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB LinearSlope function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  r = _R(dim, seed, b"R")
  z = np.matmul(r, arr)
  result = 0.0
  for i in range(dim):
    s = 10**(i / float(dim - 1) if dim > 1 else 1)
    z_opt = 5 * np.sum(np.abs(r[i, :]))
    result += float(s * (z_opt - z[i]))
  return result


def AttractiveSector(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Attractive Sector function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  x_opt = np.array([1 if i % 2 == 0 else -1 for i in range(dim)])
  x_opt.shape = (dim, 1)
  z_vec = np.matmul(_R(dim, seed, b"R"), arr - x_opt)
  z_vec = np.matmul(LambdaAlpha(10.0, dim), z_vec)
  z_vec = np.matmul(_R(dim, seed, b"Q"), z_vec)

  result = 0.0
  for i in range(dim):
    z = z_vec[i, 0]
    s = 100 if z * x_opt[i] > 0 else 1
    result += (s * z)**2

  return math.pow(Tosz(result), 0.9)


def StepEllipsoidal(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB StepEllipsoidal function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  z_hat = np.matmul(_R(dim, seed, b"R"), arr)
  z_hat = np.matmul(LambdaAlpha(10.0, dim), z_hat)
  z_tilde = np.array([
      math.floor(0.5 + z) if (z > 0.5) else (math.floor(0.5 + 10 * z) / 10)
      for z in z_hat.flat
  ])
  z_tilde = np.matmul(_R(dim, seed, b"Q"), z_tilde)
  s = 0.0
  for i, val in enumerate(z_tilde):
    exponent = 2.0 * float(i) / (dim - 1.0) if dim > 1.0 else 2.0
    s += 10.0**exponent * val**2
  value = max(abs(z_hat[0, 0]) / 1000, s)
  return 0.1 * value + Fpen(arr)


def RosenbrockRotated(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB RosenbrockRotated function."""
  dim = len(arr)
  r_x = np.matmul(_R(dim, seed, b"R"), arr)
  z = max(1.0, (dim**0.5) / 8.0) * r_x + 0.5 * np.ones((dim,))
  return float(sum(
      [100.0 * (z[i]**2 - z[i + 1])**2 + (z[i] - 1)**2
       for i in range(dim - 1)]))


def Ellipsoidal(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Ellipsoidal function."""
  del seed
  dim = len(arr)
  arr.shape = (dim, 1)
  z_vec = ArrayMap(arr, Tosz)
  s = 0.0
  for i in range(dim):
    exp = 6.0 * i / (dim - 1) if dim > 1 else 6.0
    s += float(10**exp * z_vec[i] * z_vec[i])
  return s


def Discus(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Discus function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  r_x = np.matmul(_R(dim, seed, b"R"), arr)
  z_vec = ArrayMap(r_x, Tosz)
  return float(10**6 * z_vec[0] * z_vec[0]) + sum(
      [z * z for z in z_vec[1:].flat])


def BentCigar(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB BentCigar function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  z_vec = np.matmul(_R(dim, seed, b"R"), arr)
  z_vec = Tasy(z_vec, 0.5)
  z_vec = np.matmul(_R(dim, seed, b"R"), z_vec)
  return float(z_vec[0]**2) + 10**6 * np.sum(z_vec[1:]**2)


def SharpRidge(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB SharpRidge function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  z_vec = np.matmul(_R(dim, seed, b"R"), arr)
  z_vec = np.matmul(LambdaAlpha(10, dim), z_vec)
  z_vec = np.matmul(_R(dim, seed, b"Q"), z_vec)
  return z_vec[0, 0]**2 + 100 * np.sum(z_vec[1:]**2)**0.5


def DifferentPowers(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB DifferentPowers function."""
  dim = len(arr)
  z = np.matmul(_R(dim, seed, b"R"), arr)
  s = 0.0
  for i in range(dim):
    exp = 2 + 4 * i / (dim - 1) if dim > 1 else 6
    s += abs(z[i])**exp
  return s**0.5


def Weierstrass(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Weierstrass function."""
  k_order = 12
  dim = len(arr)
  arr.shape = (dim, 1)
  z = np.matmul(_R(dim, seed, b"R"), arr)
  z = ArrayMap(z, Tosz)
  z = np.matmul(_R(dim, seed, b"Q"), z)
  z = np.matmul(LambdaAlpha(1.0 / 100.0, dim), z)
  f0 = sum([0.5**k * math.cos(math.pi * 3**k) for k in range(k_order)])

  s = 0.0
  for i in range(dim):
    for k in range(k_order):
      s += 0.5**k * math.cos(2 * math.pi * (3**k) * (z[i] + 0.5))

  return float(10 * (s / dim - f0)**3) + 10 * Fpen(arr) / dim


def SchaffersF7(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Weierstrass function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  if dim == 1:
    return 0.0
  z = np.matmul(_R(dim, seed, b"R"), arr)
  z = Tasy(z, 0.5)
  z = np.matmul(_R(dim, seed, b"Q"), z)
  z = np.matmul(LambdaAlpha(10.0, dim), z)

  s_arr = np.zeros(dim - 1)
  for i in range(dim - 1):
    s_arr[i] = float((z[i]**2 + z[i + 1]**2)**0.5)
  s = 0.0
  for i in range(dim - 1):
    s += s_arr[i]**0.5 + (s_arr[i]**0.5) * math.sin(50 * s_arr[i]**0.2)**2

  return (s / (dim - 1.0))**2 + 10 * Fpen(arr)


def SchaffersF7IllConditioned(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB SchaffersF7 Ill Conditioned."""
  dim = len(arr)
  arr.shape = (dim, 1)
  if dim == 1:
    return 0.0
  z = np.matmul(_R(dim, seed, b"R"), arr)
  z = Tasy(z, 0.5)
  z = np.matmul(_R(dim, seed, b"Q"), z)
  z = np.matmul(LambdaAlpha(1000.0, dim), z)

  s_arr = np.zeros(dim - 1)
  for i in range(dim - 1):
    s_arr[i] = float((z[i]**2 + z[i + 1]**2)**0.5)
  s = 0.0
  for i in range(dim - 1):
    s += s_arr[i]**0.5 + (s_arr[i]**0.5) * math.sin(50 * s_arr[i]**0.2)**2

  return (s / (dim - 1.0))**2 + 10 * Fpen(arr)


def GriewankRosenbrock(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB GriewankRosenbrock function."""
  dim = len(arr)
  r_x = np.matmul(_R(dim, seed, b"R"), arr)
  # Slightly off BBOB documentation in order to center optima at origin.
  # Should be: max(1.0, (dim**0.5) / 8.0) * r_x + 0.5 * np.ones((dim,)).
  z_arr = max(1.0, (dim**0.5) / 8.0) * r_x + np.ones((dim,))
  s_arr = np.zeros(dim)
  for i in range(dim - 1):
    s_arr[i] = 100.0 * (z_arr[i]**2 - z_arr[i + 1])**2 + (z_arr[i] - 1)**2

  total = 0.0
  for i in range(dim - 1):
    total += (s_arr[i] / 4000.0 - math.cos(s_arr[i]))

  return (10.0 * total) / (dim - 1) + 10


def Schwefel(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Schwefel function."""
  del seed
  dim = len(arr)
  bernoulli_arr = np.array([pow(-1, i + 1) for i in range(dim)])
  x_opt = 4.2096874633 / 2.0 * bernoulli_arr
  x_hat = 2.0 * (bernoulli_arr * arr)  # Element-wise multiplication

  z_hat = np.zeros([dim, 1])
  z_hat[0, 0] = x_hat[0]
  for i in range(1, dim):
    z_hat[i, 0] = x_hat[i] + 0.25 * (x_hat[i - 1] - 2 * abs(x_opt[i - 1]))

  x_opt.shape = (dim, 1)
  z_vec = 100 * (
      np.matmul(LambdaAlpha(10, dim), z_hat - 2 * abs(x_opt)) + 2 * abs(x_opt))

  total = sum([z * math.sin(abs(z)**0.5) for z in z_vec.flat])

  return -(total / (100.0 * dim)) + 4.189828872724339 + 100 * Fpen(z_vec / 100)


def Katsuura(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Katsuura function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  r_x = np.matmul(_R(dim, seed, b"R"), arr)
  z_vec = np.matmul(LambdaAlpha(100.0, dim), r_x)
  z_vec = np.matmul(_R(dim, seed, b"Q"), z_vec)

  prod = 1.0
  for i in range(dim):
    s = 0.0
    for j in range(1, 33):
      s += abs(2**j * z_vec[i, 0] - round(2**j * z_vec[i, 0])) / 2**j
    prod *= (1 + (i + 1) * s)**(10.0 / dim**1.2)

  return (10.0 / dim**2) * prod - 10.0 / dim**2 + Fpen(arr)


def Lunacek(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Lunacek function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  mu0 = 2.5
  s = 1.0 - 1.0 / (2.0 * (dim + 20.0)**0.5 - 8.2)
  mu1 = -((mu0**2 - 1) / s)**0.5

  x_opt = np.array([mu0 / 2] * dim)
  x_hat = np.array([2 * arr[i, 0] * np.sign(x_opt[i]) for i in range(dim)])
  x_vec = x_hat - mu0
  x_vec.shape = (dim, 1)
  x_vec = np.matmul(_R(dim, seed, b"R"), x_vec)
  z_vec = np.matmul(LambdaAlpha(100, dim), x_vec)
  z_vec = np.matmul(_R(dim, seed, b"Q"), z_vec)

  s1 = sum([(val - mu0)**2 for val in x_hat])
  s2 = sum([(val - mu1)**2 for val in x_hat])
  s3 = sum([math.cos(2 * math.pi * z) for z in z_vec.flat])
  return min(s1, dim + s * s2) + 10.0 * (dim - s3) + 10**4 * Fpen(arr)


def Gallagher101Me(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Gallagher101 function."""
  dim = len(arr)
  arr.shape = (dim, 1)

  num_optima = 101
  optima_list = [np.zeros([dim, 1])]
  for i in range(num_optima - 1):
    vec = np.zeros([dim, 1])
    for j in range(dim):
      alpha = (i * dim + j + 1.0) / (dim * num_optima + 2.0)
      assert alpha > 0
      assert alpha < 1
      vec[j, 0] = -5 + 10 * alpha
    optima_list.append(vec)

  c_list = [LambdaAlpha(1000, dim)]
  for i in range(num_optima - 1):
    alpha = 1000.0**(2.0 * (i) / (num_optima - 2))
    c_mat = LambdaAlpha(alpha, dim) / (alpha**0.25)
    c_list.append(c_mat)

  rotation = _R(dim, seed, b"R")
  max_value = -1.0
  for i in range(num_optima):
    w = 10 if i == 0 else (1.1 + 8.0 * (i - 1.0) / (num_optima - 2.0))
    diff = np.matmul(rotation, arr - optima_list[i])
    e = np.matmul(diff.transpose(), np.matmul(c_list[i], diff))
    max_value = max(max_value, w * math.exp(-float(e) / (2.0 * dim)))

  return Tosz(10.0 - max_value)**2 + Fpen(arr)


def Gallagher21Me(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Gallagher21 function."""
  dim = len(arr)
  arr.shape = (dim, 1)

  num_optima = 21
  optima_list = [np.zeros([dim, 1])]
  for i in range(num_optima - 1):
    vec = np.zeros([dim, 1])
    for j in range(dim):
      alpha = (i * dim + j + 1.0) / (dim * num_optima + 2.0)
      assert alpha > 0
      assert alpha < 1
      vec[j, 0] = -5 + 10 * alpha
    optima_list.append(vec)

  c_list = [LambdaAlpha(1000, dim)]
  for i in range(num_optima - 1):
    alpha = 1000.0**(2.0 * (i) / (num_optima - 2))
    c_mat = LambdaAlpha(alpha, dim) / (alpha**0.25)
    c_list.append(c_mat)

  rotation = _R(dim, seed, b"R")
  max_value = -1.0
  for i in range(num_optima):
    w = 10 if i == 0 else (1.1 + 8.0 * (i - 1.0) / (num_optima - 2.0))
    diff = np.matmul(rotation, arr - optima_list[i])
    e = np.matmul(diff.transpose(), np.matmul(c_list[i], diff))
    max_value = max(max_value, w * math.exp(-float(e) / (2.0 * dim)))

  return Tosz(10.0 - max_value)**2 + Fpen(arr)


## Additional BBOB-like functions to test exploration.


def NegativeSphere(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for BBOB Sphere function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  z = np.matmul(_R(dim, seed, b"R"), arr)
  return float(100 + np.sum(z * z) - 2 * (z[0]**2))


def NegativeMinDifference(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for NegativeMinDifference function."""
  dim = len(arr)
  arr.shape = (dim, 1)
  z = np.matmul(_R(dim, seed, b"R"), arr)
  min_difference = 10000
  for i in range(len(z) - 1):
    min_difference = min(min_difference, z[i + 1] - z[i])
  return 10.0 - float(min_difference) + 1e-8 * float(sum(arr))


def FonsecaFleming(arr: np.ndarray, seed: int = 0) -> float:
  """Implementation for FonsecaFleming function."""
  del seed
  return 1.0 - float(np.exp(-np.sum(arr * arr)))



"""Import target for oss."""

# TODO: Re-evalaute what to expose to users, once closed.
# TODO: Alternatively, remove this import.
from vizier._src.pyvizier.oss import metadata_util
from vizier._src.pyvizier.oss.automated_stopping import AutomatedStoppingConfig
from vizier._src.pyvizier.oss.automated_stopping import AutomatedStoppingConfigProto
from vizier._src.pyvizier.oss.proto_converters import MeasurementConverter
from vizier._src.pyvizier.oss.proto_converters import MonotypeParameterSequence
from vizier._src.pyvizier.oss.proto_converters import ParameterConfigConverter
from vizier._src.pyvizier.oss.proto_converters import ParameterType
from vizier._src.pyvizier.oss.proto_converters import ParameterValueConverter
from vizier._src.pyvizier.oss.proto_converters import ScaleType
from vizier._src.pyvizier.oss.proto_converters import TrialConverter
from vizier._src.pyvizier.oss.study_config import Algorithm
from vizier._src.pyvizier.oss.study_config import ExternalType
from vizier._src.pyvizier.oss.study_config import ObjectiveMetricGoal
from vizier._src.pyvizier.oss.study_config import ObservationNoise
from vizier._src.pyvizier.oss.study_config import ParameterValueSequence
from vizier._src.pyvizier.oss.study_config import SearchSpaceSelector
from vizier._src.pyvizier.oss.study_config import StudyConfig
from vizier._src.pyvizier.shared.base_study_config import MetricInformation
from vizier._src.pyvizier.shared.base_study_config import MetricsConfig
from vizier._src.pyvizier.shared.base_study_config import ProblemStatement
from vizier._src.pyvizier.shared.base_study_config import SearchSpace
from vizier._src.pyvizier.shared.common import Metadata
from vizier._src.pyvizier.shared.common import MetadataValue
from vizier._src.pyvizier.shared.parameter_config import ParameterConfig
from vizier._src.pyvizier.shared.trial import CompletedTrial
from vizier._src.pyvizier.shared.trial import CompletedTrialWithMeasurements
from vizier._src.pyvizier.shared.trial import Measurement
from vizier._src.pyvizier.shared.trial import Metric
from vizier._src.pyvizier.shared.trial import NaNMetric
from vizier._src.pyvizier.shared.trial import ParameterDict
from vizier._src.pyvizier.shared.trial import ParameterValue
from vizier._src.pyvizier.shared.trial import PendingTrial
from vizier._src.pyvizier.shared.trial import PendingTrialWithMeasurements
from vizier._src.pyvizier.shared.trial import Trial
from vizier._src.pyvizier.shared.trial import TrialFilter
from vizier._src.pyvizier.shared.trial import TrialStatus
from vizier._src.pyvizier.shared.trial import TrialSuggestion



"""PyVizier classes for Pythia policies."""

from vizier._src.pyvizier.pythia.study import StudyDescriptor
from vizier._src.pyvizier.shared.base_study_config import MetricInformation
from vizier._src.pyvizier.shared.base_study_config import MetricsConfig
from vizier._src.pyvizier.shared.base_study_config import MetricType
from vizier._src.pyvizier.shared.base_study_config import ObjectiveMetricGoal
from vizier._src.pyvizier.shared.base_study_config import ProblemStatement
from vizier._src.pyvizier.shared.base_study_config import SearchSpace
from vizier._src.pyvizier.shared.base_study_config import SearchSpaceSelector
from vizier._src.pyvizier.shared.common import Metadata
from vizier._src.pyvizier.shared.common import MetadataValue
from vizier._src.pyvizier.shared.common import Namespace
from vizier._src.pyvizier.shared.parameter_config import ExternalType
from vizier._src.pyvizier.shared.parameter_config import ParameterConfig
from vizier._src.pyvizier.shared.parameter_config import ParameterType
from vizier._src.pyvizier.shared.parameter_config import ScaleType
from vizier._src.pyvizier.shared.trial import CompletedTrial
from vizier._src.pyvizier.shared.trial import Measurement
from vizier._src.pyvizier.shared.trial import Metric
from vizier._src.pyvizier.shared.trial import ParameterDict
from vizier._src.pyvizier.shared.trial import ParameterValue
from vizier._src.pyvizier.shared.trial import ParameterValueTypes
from vizier._src.pyvizier.shared.trial import Trial
from vizier._src.pyvizier.shared.trial import TrialFilter
from vizier._src.pyvizier.shared.trial import TrialStatus
from vizier._src.pyvizier.shared.trial import TrialSuggestion

FlatSpace = SearchSpace
StudyConfig = ProblemStatement  # To be deprecated.



"""Abstractions."""

import abc
from typing import Optional, Sequence, TypeVar

import attr
from vizier import pyvizier as vz
from vizier.interfaces import serializable

_T = TypeVar('_T')


@attr.define(frozen=True)
class CompletedTrials:
  """A group of completed trials.

  Attributes:
    completed: Completed Trials.
  """

  def __attrs_post_init__(self):
    for trial in self.completed:
      if trial.status != vz.TrialStatus.COMPLETED:
        raise ValueError(f'All trials must be completed. Bad trial:\n{trial}')

  completed: Sequence[vz.Trial] = attr.field(
      converter=tuple,
      validator=attr.validators.deep_iterable(
          attr.validators.instance_of(vz.Trial)))


class _SuggestionAlgorithm(abc.ABC):
  """Suggestion algorithm."""

  @abc.abstractmethod
  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    pass


class Designer(_SuggestionAlgorithm):
  """Suggestion algorithm for sequential usage.

  `Designer` is the recommended interface for implementing commonly used
  algorithms such as GP-UCB and evolutionary algorithms. A `Designer` can be
  wrapped into a pythia `Policy` via `DesignerPolicy`. When run inside a service
  binary, `Designer` instance is not guaranteed to persist during
  the lifetime of a `Study`, and should be assumed to receive all trials from
  the beginning of the study in `update()` calls.

  If your `Designer` can benefit from a persistent state, implement a
  either `SerializableDesigner` or `PartiallySerializableDesigner`, and use
  `SerializableDesignerPolicy` or `PartiallySerializableDesignerPolicy`,
  respectively.
  """

  @abc.abstractmethod
  def update(self, delta: CompletedTrials) -> None:
    """Reflect the delta in the designer's state."""
    pass


class PartiallySerializableDesigner(Designer,
                                    serializable.PartiallySerializable):
  pass


class SerializableDesigner(Designer, serializable.Serializable):
  pass



"""The Policy can use these classes to communicate with Vizier."""

import abc
import collections
import dataclasses
import datetime
from typing import Dict, Iterable, List, Optional

from vizier import pyvizier as vz


@dataclasses.dataclass(frozen=True)
class MetadataDelta:
  """Carries cumulative delta for a batch metadata update.

  Attributes:
    on_study: Updates to be made on study-level metadata.
    on_trials: Maps trial id to updates.
  """

  on_study: vz.Metadata = dataclasses.field(default_factory=vz.Metadata)
  on_trials: Dict[int, vz.Metadata] = dataclasses.field(
      default_factory=lambda: collections.defaultdict(vz.Metadata))


class _MetadataUpdateContext:
  """Metadata update context.

  Usage:
    # All metadata updates in the context are queued, not immediately applied.
    # Upon exit, supporter handles all metadata updates in a batch.
    with pythia2._MetadataUpdateContext(policy_supporter) as mu:
      # Study-level metadata.
      mu.assign('namespace', 'key', 'value')
      # Trial-level metadata.
      mu.assign('namespace', 'key', 'value', trial_id=1)
      # Same as above but with a side effect. After this line the following
      # line is True:
      #   trial.metadata.ns('namespace')['key'] == 'value'
      mu.assign('namespace', 'key', 'value', trial)
  """

  def __init__(self, supporter: 'PolicySupporter'):
    self._supporter = supporter
    self._delta = MetadataDelta()

  # pylint: disable=invalid-name
  def assign(self,
             namespace: str,
             key: str,
             value: vz.MetadataValue,
             trial: Optional[vz.Trial] = None,
             *,
             trial_id: Optional[int] = None):
    """Assigns metadata.

    Args:
      namespace: Namespace of the metadata. See vz.Metadata doc for more
        details.
      key:
      value:
      trial: If specified, `trial_id` must be None. It behaves the same as when
        `trial_id=trial.id`, except that `trial` is immediately modified.
      trial_id: If specified, `trial` must be None. If both `trial` and
        `trial_id` are None, then the key-value pair will be assigned to the
        study.

    Raises:
      ValueError:
    """
    if trial is None and trial_id is None:
      self._delta.on_study.ns(namespace)[key] = value
    elif trial is not None and trial_id is not None:
      raise ValueError(
          'At most one of `trial` and `trial_id` can be specified.')
    elif trial is not None:
      self._delta.on_trials[trial.id].ns(namespace)[key] = value
      trial.metadata.ns(namespace)[key] = value
    elif trial_id is not None:
      self._delta.on_trials[trial_id].ns(namespace)[key] = value

  def __enter__(self):
    return self

  def __exit__(self, *args):
    """upon exit, sends a batch update request."""
    self._supporter.SendMetadata(self._delta)


class PolicySupporter(abc.ABC):
  """Used by Policy instances to communicate with Vizier."""

  # TODO: Change to GetStudyDescriptor.
  @abc.abstractmethod
  def GetStudyConfig(self, study_guid: Optional[str] = None) -> vz.StudyConfig:
    """Requests a StudyConfig from Vizier.

    This sends a PythiaToVizier.trial_selector packet and waits for the
    response(s).  You can call this multiple times, and it is thread-friendly,
    so you can even overlap calls.

    Args:
      study_guid: The GUID of the study whose StudyConfig you want. Note that
        access control applies. By default, use the current study's GUID.

    Returns:
      The requested StudyConfig proto.

    Raises:
      CancelComputeError: (Do not catch.)
      PythiaProtocolError: (Do not catch.)
      VizierDatabaseError: If the database operation raises an error, e.g. if
        $study_guid refers to a nonexistent or inaccessible study.
    """

  @abc.abstractmethod
  def GetTrials(
      self,
      *,
      study_guid: Optional[str] = None,
      trial_ids: Optional[Iterable[int]] = None,
      min_trial_id: Optional[int] = None,
      max_trial_id: Optional[int] = None,
      status_matches: Optional[vz.TrialStatus] = None,
      include_intermediate_measurements: bool = True) -> List[vz.Trial]:
    """Requests Trials from Vizier.

    Args:
      study_guid: The GUID of the study to get Trials from.  Default is None,
        which means the current Study.
      trial_ids: a list of Trial id numbers to acquire.
      min_trial_id: Trials in [min_trial_id, max_trial_id] are selected, if at
        least one of the two is not None.
      max_trial_id: Trials in [min_trial_id, max_trial_id] are selected, if at
        least one of the two is not None.
      status_matches: If not None, filters for Trials where
        Trial.status==status_matches.  The default passes all types of Trial.
      include_intermediate_measurements: If True (default), the returned Trials
        must have all measurements. Note that the final Measurement is always
        included for COMPLETED Trials. If False, PolicySupporter _may_ leave
        `measurements` field empty in the returned Trials in order to optimize
        speed, but it is not required to do so.

    Returns:
      Trials obtained from Vizier.

    Raises:
      CancelComputeError: (Do not catch.)
      PythiaProtocolError: (Do not catch.)
      VizierDatabaseError: If the database operation raises an error, e.g. if
        $study_guid refers to a nonexistent or inaccessible study.

    NOTE: if $trial_ids is set, $min_trial_id, $max_trial_id, and
      $status_matches will be ignored.
    """

  def CheckCancelled(self, note: Optional[str] = None) -> None:
    """Throws a CancelComputeError on timeout or if Vizier cancels.

    This should be called occasionally by any long-running computation.
    Raises an exception if the interaction has been cancelled by the Vizier
    side of the protocol; the exception shuts down the Pythia server.

    Args:
      note: for debugging.

    Raises:
      CancelComputeError: (Do not catch.)
    """
    pass

  def TimeRemaining(self) -> datetime.timedelta:
    """The time remaining to compute a result.

    Returns:
      The remaining time before the RPC is considered to have timed out; it
      returns datetime.timedelta.max if no deadline was specified in the RPC.

    This is an alternative to calling CheckCancelled(); both have the goal of
    terminating runaway computations.  If your computation times out,
    you should raise TemporaryPythiaError (if you want a retry) or
    InactivateStudyError (if not).
    """
    return datetime.timedelta(hours=1.0)

  def MetadataUpdate(self) -> _MetadataUpdateContext:
    """Queues metadata updates, then passes them to UpdateMetadata().

    Usage:
      ps = PolicySupporter()
      with ps.MetadataUpdate() as mu:
        # Study-level metadata.
        mu.assign('namespace', 'key', 'value')
        # Trial-level metadata.
        mu.assign('namespace', 'key', 'value', trial_id=1)

    Returns:
      A _MetadataUpdateContext instance to use as a context.
    Raises:
      CancelComputeError: (Do not catch.)
      PythiaProtocolError: (Do not catch.)
      VizierDatabaseError: If the database operation raises an error.
    """
    return _MetadataUpdateContext(self)

  @abc.abstractmethod
  def SendMetadata(self, delta: MetadataDelta) -> None:
    """Updates the Study's metadata in Vizier's database.

    The MetadataUpdate() method is preferred for normal use.

    Args:
      delta: Metadata to be uploaded to the Vizier database.

    Raises:
      CancelComputeError: (Do not catch.)
      PythiaProtocolError: (Do not catch.)
      VizierDatabaseError: If the database operation raises an error.
    """



"""Abstractions and default converters."""

import abc
import copy
import dataclasses
import enum
import itertools
from typing import Any, Callable, Collection, Dict, Iterator, List, Mapping, Optional, Sequence, Tuple, Union

from absl import logging
import attr
import numpy as np

from vizier import pyvizier

# The study identifier for cross-study learning must be stored in
# Trial.Metadata and StudyConfig.Metadata with this key.
# TODO: Use metadata namespace instead.
STUDY_ID_FIELD = 'metalearn_study_id'


class NumpyArraySpecType(enum.Enum):
  """Type information for NumpyArraySpec.

  CONTINUOUS: Continuous parameter
  DISCRETE: Discrete/integer/categorical parameter
  ONEHOT_EMBEDDING: One-hot embedding of DISCRETE.
  """
  CONTINUOUS = 'CONTINUOUS'
  DISCRETE = 'DISCRETE'
  ONEHOT_EMBEDDING = 'ONEHOT_EMBEDDING'

  @classmethod
  def default_factory(cls,
                      pc: pyvizier.ParameterConfig) -> 'NumpyArraySpecType':
    """SpecType when encoding discretes as integer indices."""
    if pc.type == pyvizier.ParameterType.DOUBLE:
      return NumpyArraySpecType.CONTINUOUS
    elif pc.type in (pyvizier.ParameterType.DISCRETE,
                     pyvizier.ParameterType.CATEGORICAL,
                     pyvizier.ParameterType.INTEGER):
      return NumpyArraySpecType.DISCRETE
    raise ValueError(f'Unknown type {pc.type}')

  @classmethod
  def embedding_factory(cls,
                        pc: pyvizier.ParameterConfig) -> 'NumpyArraySpecType':
    """SpecType when encoding discretes as onehot embedding."""
    if pc.type == pyvizier.ParameterType.DOUBLE:
      return NumpyArraySpecType.CONTINUOUS
    elif pc.type in (pyvizier.ParameterType.DISCRETE,
                     pyvizier.ParameterType.CATEGORICAL,
                     pyvizier.ParameterType.INTEGER):
      return NumpyArraySpecType.ONEHOT_EMBEDDING
    raise ValueError(f'Unknown type {pc.type}')


@attr.define(frozen=True, auto_attribs=True)
class NumpyArraySpec:
  """Encodes what an array represents.

  This class is similar to `BoundedTensorSpec` in tf agents, except it carries
  extra information specific to vizier.

  If `type` is `DOUBLE`, then `dtype` is a floating type, and bounds are
  floating numbers. num_dimensions is always 1, and num_oovs is zero.

  If 'type' is `DISCRETE`, then `dtype` is an integer type, and bounds are
  integers. num_dimensions is always 1. Suppose `bounds=(x,y)`. Then integers
  x to (y-num_oovs) correspond to valid parameter values. The rest represent
  out-of-vocabulary values. For example, an integer parameter in range (1,3)
  can be represented by a DISCRETE NumpyArraySpec with bounds=(1,4) and oov=1.

  If 'type' is `ONEHOT_EMBEDDING`, then `dtype` is a floating type, and bounds
  are floating numbers. Suppose num_dimensions is X.

  Attributes:
    type: Underlying type of the Vizier parameter corresponding to the array.
    dtype: Numpy array's type.
    bounds: Always inclusive in both directions.
    num_dimensions: Corresponds to shape[-1] of the numpy array. When `type` is
      `ONEHOT_EMBEDDING`, the first X dimensions correspond to valid parameter
      values. The other dimensions correspond to out-of-vocabulary values.
      Otherwise, it is simply 1.
    name: Parameter name.
    num_oovs: Number of out-of-vocabulary items, for non-continuous type.
    scale: Scaling of the values.
  """
  type: NumpyArraySpecType
  dtype: np.dtype
  bounds: Union[Tuple[float, float], Tuple[int, int]]
  num_dimensions: int
  name: str
  num_oovs: int
  scale: Optional[pyvizier.ScaleType] = None

  def __attrs_post_init__(self):
    object.__setattr__(self, 'bounds',
                       tuple(np.array(self.bounds, dtype=self.dtype)))

  @classmethod
  def from_parameter_config(
      cls,
      pc: pyvizier.ParameterConfig,
      type_factory: Callable[
          [pyvizier.ParameterConfig],
          NumpyArraySpecType] = NumpyArraySpecType.default_factory,
      floating_dtype: np.dtype = np.float32,
      int_dtype: np.dtype = np.int32,
      *,
      pad_oovs: bool = True) -> 'NumpyArraySpec':
    """Factory function.

    Args:
      pc:
      type_factory: NumpyArraySpecType has `default_factory` and
        `embedding_factory`. The difference is in how they handle non-continuous
        parameters.
      floating_dtype: Dtype of the floating outputs.
      int_dtype: Dtype of the integer outputs.
      pad_oovs: If True, pad the out-of-vocabulary dimensions to onehot
        embedding.

    Returns:
      NumpyArraySpec.
    """
    the_type = type_factory(pc)
    if the_type == NumpyArraySpecType.CONTINUOUS:
      return NumpyArraySpec(
          the_type,
          floating_dtype,
          bounds=pc.bounds,
          num_dimensions=1,
          scale=pc.scale_type,
          name=pc.name,
          num_oovs=0)
    elif the_type == NumpyArraySpecType.DISCRETE:
      return NumpyArraySpec(
          the_type,
          int_dtype,
          bounds=(0, len(pc.feasible_values)),
          num_dimensions=1,
          name=pc.name,
          num_oovs=1 if pad_oovs else 0)
    elif the_type == NumpyArraySpecType.ONEHOT_EMBEDDING:
      return NumpyArraySpec(
          the_type,
          floating_dtype,
          bounds=(0., 1.),
          num_dimensions=len(pc.feasible_values) + 1,
          name=pc.name,
          num_oovs=1 if pad_oovs else 0)
    raise ValueError(f'Unknown type {type}')


def dict_to_array(array_dict: Dict[str, np.ndarray]) -> np.ndarray:
  r"""Converts a dict of (..., D_i) arrays to a (..., \sum_i D_i) array."""
  return np.concatenate(list(array_dict.values()), axis=-1)


class DictOf2DArrays(Mapping[str, np.ndarray]):
  """Dictionary of string to 2D arrays.

  All arrays share the first dimension, which is at a high level, the number of
  objects that this dictionary corresponds to.

  Attributes:
    size: Array's shape[0].
  """

  def __init__(self, d: Dict[str, np.ndarray]):
    self._d = d
    shape = None
    for k, v in self.items():
      if shape is None:
        shape = v.shape
        if len(shape) != 2:
          raise ValueError(f'{k} has shape {v.shape} which is not length 2.'
                           'DictOf2DArrays only supports 2D numpy arrays.')
      if shape[0] != v.shape[0]:
        raise ValueError(
            f'{k} has shape {v.shape} which is not equal to {shape}.')
    self._size = shape[0]

  def __getitem__(self, key: str) -> np.ndarray:
    return self._d[key]

  def __iter__(self) -> Iterator[str]:
    return iter(self._d)

  def __len__(self) -> int:
    return len(self._d)

  def __add__(self, other: 'DictOf2DArrays') -> 'DictOf2DArrays':
    if not isinstance(other, DictOf2DArrays):
      raise ValueError('You can add DictOf2DArrays only.')
    if len(self) != len(other):
      # We don't check the keys because it's too slow.
      raise ValueError('Two arrays have different length!')
    return DictOf2DArrays(
        {k: np.concatenate([self[k], other[k]], axis=0) for k in self})

  @property
  def size(self) -> int:
    return self._size

  def asarray(self) -> np.ndarray:
    return dict_to_array(self._d)

  def dict_like(self, array: np.ndarray) -> 'DictOf2DArrays':
    """[Experimental] Converts an array into a dict with the same keys as this.

    This function acts like an inverse of `asarray()`, i.e. it satisfies
      `self.dict_like(self.asarray()) == self`.

    Example:
      d = DictOf2DArrays({'p1': [[1], [2], [3]], 'p2': [[4], [5], [6]]})
      d.dict_like([[1, 2], [3, 4]]) == {'p1': [[1], [3]], 'p2': [[2],[4]]}

    Args:
      array:

    Returns:
      DictOf2DArrays with the same shape spec as this.
    """
    begin = 0
    new_dict = dict()
    for k, v in self.items():
      end = begin + v.shape[1]
      new_dict[k] = array[:, begin:end].astype(v.dtype)
      begin = end

    return DictOf2DArrays(new_dict)


class TrialToNumpyDict(abc.ABC):
  """Parses a sequence of Trials to a dict keyed by parameter and metric names.

  A typical Keras/JAX pipeline consists of:
    1. Load data into arrays.
    2. Call Model.build() to initialize a model for the loaded data shape.
    3. Call Model.fit() to train the model.
    4. Call Model.__call__() to predict with the model.

  This abstraction allows a shared implementation of steps 1,2 and 3.
  """

  @abc.abstractmethod
  def to_xy(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
    """Returns (x,y) pair that can be used as input for keras.Model.fit()."""
    pass

  @property
  @abc.abstractmethod
  def features_shape(self) -> Dict[str, Sequence[Union[int, None]]]:
    """Returned value can be used as `input_shape` for keras.Model.build()."""
    pass

  @property
  @abc.abstractmethod
  def output_specs(self) -> Dict[str, NumpyArraySpec]:
    """Same keys as features_shape, with more details."""
    pass

  @property
  @abc.abstractmethod
  def labels_shape(self) -> Dict[str, Any]:
    pass

  @property
  @abc.abstractmethod
  def metric_information(self) -> Dict[str, pyvizier.MetricInformation]:
    pass


class ModelInputConverter(metaclass=abc.ABCMeta):
  """Interface for extracting inputs to the model."""

  @abc.abstractmethod
  def convert(self, trials: Sequence[pyvizier.Trial]) -> np.ndarray:
    """Returns an array of shape (number of trials, feature dimension).

    Args:
      trials:

    Returns:
      Returns an array of shape (number of trials, feature dimension).
      Subclasses must use a fixed feature dimension. In particular, it should
      be a constant function of the input trials.
    """
    pass

  @property
  @abc.abstractmethod
  def output_spec(self) -> NumpyArraySpec:
    """Provides specification of the output from this converter."""
    pass

  @property
  @abc.abstractmethod
  def parameter_config(self):
    """Original ParameterConfig that this converter acts on."""
    pass

  @abc.abstractmethod
  def to_parameter_values(
      self, array: np.ndarray) -> List[Optional[pyvizier.ParameterValue]]:
    """Convert to parameter values."""
    pass


@dataclasses.dataclass
class _ModelInputArrayBijector:
  """Transformations on the numpy arrays generated by ModelInputConverter."""
  forward_fn: Callable[[np.ndarray], np.ndarray]
  backward_fn: Callable[[np.ndarray], np.ndarray]
  output_spec: NumpyArraySpec  # Spec after forward_fn is applied.

  @classmethod
  def identity(cls, spec) -> '_ModelInputArrayBijector':
    return cls(lambda x: x, lambda x: x, spec)

  @classmethod
  def scaler_from_spec(cls, spec: NumpyArraySpec) -> '_ModelInputArrayBijector':
    """For continuous specs, linearize and scale it to (0, 1) range."""
    low, high = spec.bounds
    if spec.type != NumpyArraySpecType.CONTINUOUS:
      return cls.identity(attr.evolve(spec, scale=None))
    if low == high:

      def backward_fn(y):
        return np.where(np.isfinite(y), np.zeros_like(y) + low, y)

      return cls(lambda x: np.where(np.isfinite(x), np.zeros_like(x), x),
                 backward_fn, attr.evolve(spec, bounds=(.0, 1.), scale=None))

    if spec.scale == pyvizier.ScaleType.LOG:
      low, high = np.log(low), np.log(high)
      denom = (high - low) or 1.0
      if denom < 1e-6:
        logging.warning('Unusually small range detected for %s', spec)
      scale_fn = lambda x, high=high, low=low: (np.log(x) - low) / (high - low)
      unscale_fn = lambda x, high=high, low=low: np.exp(x * (high - low) + low)
    else:
      if not (spec.scale == pyvizier.ScaleType.LINEAR or spec.scale is None):
        logging.warning('Unknown scale type %s. Applying LINEAR', spec.scale)
      denom = (high - low)
      if denom < 1e-6:
        logging.warning('Unusually small range detected for %s', spec)
      if denom == 1.0 and low == 0:
        return cls.identity(attr.evolve(spec, scale=None))
      scale_fn = lambda x, high=high, low=low: (x - low) / (high - low)
      unscale_fn = lambda x, high=high, low=low: x * (high - low) + low

    return cls(scale_fn, unscale_fn,
               attr.evolve(spec, bounds=(.0, .1), scale=None))

  @classmethod
  def onehot_embedder_from_spec(cls,
                                spec: NumpyArraySpec,
                                *,
                                dtype=np.float32,
                                pad_oovs: bool = True
                               ) -> '_ModelInputArrayBijector':
    """Given a discrete spec, one-hot embeds it."""
    if spec.type != NumpyArraySpecType.DISCRETE:
      return cls.identity(spec)

    num_oovs = 1 if pad_oovs else 0
    output_spec = NumpyArraySpec(
        NumpyArraySpecType.ONEHOT_EMBEDDING,
        dtype,
        bounds=(0., 1.),
        num_dimensions=int(spec.bounds[1] - spec.bounds[0] + num_oovs),
        name=spec.name,
        num_oovs=num_oovs,
        scale=None)

    def embed_fn(x: np.ndarray, output_spec=output_spec):
      """x is integer array of [N, 1]."""
      return np.eye(
          output_spec.num_dimensions, dtype=output_spec.dtype)[x.flatten()]

    def unembed_fn(x: np.ndarray, spec=spec, output_spec=output_spec):
      return np.argmax(
          x[:, :output_spec.num_dimensions - output_spec.num_oovs],
          axis=1).astype(spec.dtype)

    return cls(embed_fn, unembed_fn, output_spec)


def _create_default_getter(
    pconfig: pyvizier.ParameterConfig) -> Callable[[pyvizier.Trial], Any]:
  """Create a default getter for the given parameter config."""

  def getter(trial, pconfig=pconfig):
    if pconfig.name not in trial.parameters:
      return None

    pvalue = trial.parameters[pconfig.name]
    if pconfig.type == pyvizier.ParameterType.DOUBLE:
      return pvalue.as_float
    elif pconfig.type == pyvizier.ParameterType.DISCRETE:
      return pvalue.as_float
    elif pconfig.type == pyvizier.ParameterType.INTEGER:
      return pvalue.as_int
    else:
      return pvalue.as_str

  return getter


class DefaultModelInputConverter(ModelInputConverter):
  """Converts trials into a (None, 1) array corresponding to a parameter.


  If the parameter_config is continuous, values obtained from `getter()` are
  directly returned as floating numbers. Otherwise, this converter returns
  the index of the value obtained from `getter()` within
  `parameter_config.feasible_points` as int32.
  """

  def __init__(self,
               parameter_config: pyvizier.ParameterConfig,
               getter: Optional[Callable[[pyvizier.Trial], Any]] = None,
               *,
               float_dtype: np.dtype = np.float32,
               max_discrete_indices: int = 10,
               scale: bool = False,
               onehot_embed: bool = False,
               converts_to_parameter: bool = True,
               pad_oovs: bool = True):
    """Init.

    Given B trials, convert() always converts to (B, 1) array. The returned
    array may contain NaNs.

    Args:
      parameter_config:
      getter: See class pydoc. If the getter is not specified, the default
        getter looks up `parameter_config.name` inside `Trial.parameters`.
      float_dtype: floating precision to be used.
      max_discrete_indices: If the parameter config has more than this many
        number of DISCRETE/INTEGER feasible points, then the parameter config is
        continuified first.
      scale:
      onehot_embed:
      converts_to_parameter: If False, this converter does not correspodn to an
        actual parameter in Vizier search space, and `to_parameter_value` always
        returns None
      pad_oovs: If True, pad the out-of-vocabulary dimensions to onehot
        embedding.
    """
    self._converts_to_parameter = converts_to_parameter
    self._parameter_config = copy.deepcopy(parameter_config)
    if parameter_config.type in (
        pyvizier.ParameterType.INTEGER, pyvizier.ParameterType.DISCRETE
    ) and parameter_config.num_feasible_values > max_discrete_indices:
      parameter_config = parameter_config.continuify()

    # TODO: Make the default getter raise an Error if they encounter an
    # out-of-vocabulary value but pad_oovs is False.
    self._getter = getter or _create_default_getter(parameter_config)
    # Getter spec can only have DISCRETE or CONTINUOUS types.
    self._getter_spec = NumpyArraySpec.from_parameter_config(
        parameter_config,
        NumpyArraySpecType.default_factory,
        floating_dtype=float_dtype)

    # Optionally scale and onehot embed.
    spec = self._getter_spec
    self.scaler = (
        _ModelInputArrayBijector.scaler_from_spec(spec)
        if scale else _ModelInputArrayBijector.identity(spec))
    spec = self.scaler.output_spec
    self.onehot_encoder = (
        _ModelInputArrayBijector.onehot_embedder_from_spec(
            spec, dtype=float_dtype, pad_oovs=pad_oovs)
        if onehot_embed else _ModelInputArrayBijector.identity(spec))
    spec = self.onehot_encoder.output_spec

    self._output_spec = spec

  def convert(self, trials: Sequence[pyvizier.Trial]) -> np.ndarray:
    """Returns an array of shape [len(trials), output_spec.num_dimensions].

    Args:
      trials:

    Returns:
      For each `trial`, if `self.getter(trial)` returns `None`, we _impute_ the
      value; otherwise, we _extract_ the value.
      If `self.parameter_config.type` is `DOUBLE`, then
        * EXTRACT: Directly use the getter's return value as float.
        * IMPUTE: Return `nan`.
      Otherwise,
        * EXTRACT: Returns the integer index of the getter's return value within
          feasible values.
        * IMPUTE: Returns `len(feasible_values)`.
    """
    if not trials:
      return np.zeros([0, self.output_spec.num_dimensions],
                      dtype=self.output_spec.dtype)

    value_converter = (
        self._convert_index if self._getter_spec.type
        == NumpyArraySpecType.DISCRETE else self._convert_continuous)
    values = [value_converter(t) for t in trials]
    array = np.asarray(values, dtype=self._getter_spec.dtype).reshape([-1, 1])
    return self.onehot_encoder.forward_fn(self.scaler.forward_fn(array))

  def _to_parameter_value(
      self, value: Union['np.float', float,
                         int]) -> Optional[pyvizier.ParameterValue]:
    """Converts to a single parameter value.

    Be aware that the value is automatically truncated.

    Args:
      value:

    Returns:
      ParameterValue.
    """
    if not self._converts_to_parameter:
      return None
    if self.parameter_config.type == pyvizier.ParameterType.DOUBLE:
      # Input parameter was DOUBLE. Output is also DOUBLE.
      return pyvizier.ParameterValue(
          float(
              np.clip(value, self._parameter_config.bounds[0],
                      self._parameter_config.bounds[1])))
    elif self.output_spec.type == NumpyArraySpecType.CONTINUOUS:
      # The parameter config is originally discrete, but continuified.
      # Round to the closest number.
      return pyvizier.ParameterValue(
          min(self.parameter_config.feasible_values,
              key=lambda feasible_value: abs(feasible_value - value)))
    elif value >= len(self.parameter_config.feasible_values):
      return None
    else:
      return pyvizier.ParameterValue(
          self.parameter_config.feasible_values[value])

  def to_parameter_values(
      self, array: np.ndarray) -> List[Optional[pyvizier.ParameterValue]]:
    """Convert and clip to the nearest feasible parameter values."""
    array = self.scaler.backward_fn(self.onehot_encoder.backward_fn(array))
    return [self._to_parameter_value(v) for v in list(array.flatten())]

  def _convert_index(self, trial: pyvizier.Trial):
    """Used for non-continuous types."""
    raw_value = self._getter(trial)
    if raw_value in self.parameter_config.feasible_values:
      return self.parameter_config.feasible_values.index(raw_value)
    else:
      # Return the catch-all missing index.
      return len(self.parameter_config.feasible_values)

  def _convert_continuous(self, trial: pyvizier.Trial):
    """Used for continuous types."""
    raw_value = self._getter(trial)
    if raw_value is None:
      return np.nan
    else:
      return raw_value

  @property
  def dtype(self):
    """dtype of the array returned by convert()."""
    return self.output_spec.dtype

  @property
  def output_spec(self) -> NumpyArraySpec:
    return self._output_spec

  @property
  def parameter_config(self) -> pyvizier.ParameterConfig:
    """Original parameter config that this converter is defined on."""
    return self._parameter_config


class ModelOutputConverter(metaclass=abc.ABCMeta):
  """Metric converter interface.

  Unlike ModelInputConverter, this class is currently designed to always
  output a single-dimensional metric. This is because in standard Keras
  workflows, output shapes must be known when defining a model. (In contrast,
  input shapes are required at the time of _building_ a model.)
  """

  @abc.abstractmethod
  def convert(self, measurements: Sequence[pyvizier.Measurement]) -> np.ndarray:
    """Returns N x 1 array."""
    pass

  @property
  @abc.abstractmethod
  def metric_information(self) -> pyvizier.MetricInformation:
    """Reflects how the converter treates the metric.

    If this metric converter flips the signs or changes the semantics of
    safety configs, then the returned metric_information should reflect such
    changes.
    """
    pass

  @property
  def output_shape(self) -> Tuple[None, int]:
    return (None, 1)


class DefaultModelOutputConverter(ModelOutputConverter):
  """Converts measurements into numpy arrays."""

  def __init__(self,
               metric_information: pyvizier.MetricInformation,
               *,
               flip_sign_for_minimization_metrics: bool = False,
               shift_safe_metrics: bool = True,
               dtype: np.dtype = np.float32,
               raise_errors_for_missing_metrics: bool = False):
    """Init.

    Args:
      metric_information:
      flip_sign_for_minimization_metrics: Flips the sign if the metric is to
        minimize.
      shift_safe_metrics: If True, add (minimize) or subtract (maximize) safety
        threshold from the metric value.
      dtype:
      raise_errors_for_missing_metrics: If True, raise errors. Otherwise, fill
        with NaN.
    """

    self._original_metric_information = metric_information
    self.flip_sign_for_minimization_metrics = flip_sign_for_minimization_metrics
    self.raise_errors_for_missing_metrics = raise_errors_for_missing_metrics
    self.dtype = dtype
    self.shift_safe_metrics = shift_safe_metrics

  @property
  def _should_flip_sign(self) -> bool:
    return (self._original_metric_information.goal
            == pyvizier.ObjectiveMetricGoal.MINIMIZE and
            self.flip_sign_for_minimization_metrics)

  def convert(
      self,
      measurements: Sequence[Optional[pyvizier.Measurement]]) -> np.ndarray:
    """Returns a (len(measurements), 1) array."""
    if not measurements:
      return np.zeros([0, 1], dtype=self.dtype)

    all_metrics = [m.metrics if m is not None else dict() for m in measurements]
    if not self.raise_errors_for_missing_metrics:
      metricvalues = [
          metrics.get(self._original_metric_information.name, None)
          for metrics in all_metrics
      ]
      labels = [mv.value if mv else np.nan for mv in metricvalues]
    else:
      labels = [
          metrics[self._original_metric_information.name].value
          for metrics in all_metrics
      ]
    labels = np.asarray(labels, dtype=self.dtype)[:, np.newaxis]
    if (self.shift_safe_metrics and
        self._original_metric_information.type.is_safety):
      labels -= self._original_metric_information.safety_threshold
    return labels * (-1 if self._should_flip_sign else 1)

  @property
  def metric_information(self) -> pyvizier.MetricInformation:
    """Returns a copy that reflects how the converter treates the metric."""
    metric_information = copy.deepcopy(self._original_metric_information)
    if self._should_flip_sign:
      metric_information = metric_information.flip_goal()
    return metric_information


class DefaultTrialConverter(TrialToNumpyDict):
  """Combines parameter and metric converters.

  Always takes the final measurement and convert to metrics.

  WARNING: This class is not responsible for filtering out non-completed or
  infeasible trials. Labels may contain NaN in such cases.

  For standard use cases, use one of the factory functions. For more
  flexibility, inject parameter_converters and metric_converters on your own.
  """

  def __init__(self,
               parameter_converters: Collection[ModelInputConverter],
               metric_converters: Collection[ModelOutputConverter] = tuple()):
    self.parameter_converters = list(parameter_converters)
    self._parameter_converters_dict = {
        pc.parameter_config.name: pc for pc in self.parameter_converters
    }
    self.metric_converters = list(metric_converters)

  def to_features(self,
                  trials: Sequence[pyvizier.Trial]) -> Dict[str, np.ndarray]:
    """Returned value can be used as `x` for keras.Model.fit()."""
    result_dict = dict()
    for converter in self.parameter_converters:
      result_dict[converter.parameter_config.name] = converter.convert(trials)
    return result_dict

  def to_trials(self, dictionary: Mapping[str,
                                          np.ndarray]) -> List[pyvizier.Trial]:
    """Inverse of `to_features`."""
    return [
        pyvizier.Trial(parameters=p) for p in self.to_parameters(dictionary)
    ]

  def to_parameters(
      self, dictionary: Mapping[str,
                                np.ndarray]) -> List[pyvizier.ParameterDict]:
    """Convert to nearest feasible parameter value. NaNs are preserved."""
    # TODO: Add a boolean flag to disable automatic clipping.
    param_dicts = [
        pyvizier.ParameterDict()
        for _ in range(len(list(dictionary.values())[0]))
    ]
    for key, values in dictionary.items():
      parameter_converter = self._parameter_converters_dict[key]
      parameter_values = parameter_converter.to_parameter_values(values)
      for param_dict, value in zip(param_dicts, parameter_values):
        if value is not None:
          param_dict[key] = value
    return param_dicts

  def to_labels(self,
                trials: Sequence[pyvizier.Trial]) -> Dict[str, np.ndarray]:
    """Returned value can be used as `y` for keras.Model.fit()."""
    result_dict = dict()
    for converter in self.metric_converters:
      result_dict[converter.metric_information.name] = converter.convert(
          [t.final_measurement for t in trials])
    return result_dict

  def to_labels_array(self, trials: Sequence[pyvizier.Trial]) -> np.ndarray:
    """Shorthand for dict_to_array(self.to_labels(trials))."""
    return dict_to_array(self.to_labels(trials))

  def to_xy(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
    """Returned value can be used as `x`, `y` for keras.Model.fit()."""
    return self.to_features(trials), self.to_labels(trials)

  @property
  def features_shape(self) -> Dict[str, Tuple[Union[int, None], int]]:
    """Returned value can be used as `input_shape` for keras.Model.build()."""
    return {
        pc.output_spec.name: (None, pc.output_spec.num_dimensions)
        for pc in self.parameter_converters
    }

  def compute_features_shape(
      self, trials: Sequence[pyvizier.Trial]) -> Dict[str, Tuple[int, int]]:
    return {k: (len(trials), v[1]) for k, v in self.features_shape.items()}

  @property
  def output_specs(self) -> Dict[str, NumpyArraySpec]:
    return {
        pc.output_spec.name: pc.output_spec for pc in self.parameter_converters
    }

  @property
  def labels_shape(self) -> Dict[str, Sequence[Union[int, None]]]:
    return {
        mc.metric_information.name: mc.output_shape
        for mc in self.metric_converters
    }

  @property
  def metric_information(self) -> Dict[str, pyvizier.MetricInformation]:
    return {
        mc.metric_information.name: mc.metric_information
        for mc in self.metric_converters
    }

  @property
  def parameter_configs(self) -> Dict[str, pyvizier.ParameterConfig]:
    return {
        converter.parameter_config.name: converter.parameter_config
        for converter in self.parameter_converters
    }

  @classmethod
  def from_study_configs(
      cls,
      study_configs: Sequence[pyvizier.StudyConfig],
      metric_information: Collection[pyvizier.MetricInformation],
      *,
      use_study_id_feature: bool = True) -> 'DefaultTrialConverter':
    """Creates a converter from a list of study configs.

    Args:
      study_configs: StudyConfigs to be merged.
      metric_information: MetricInformation of metrics to be used as y-values.
      use_study_id_feature: If True, an extra parameter is added that
        corresponds to the STUDY_ID_FIELD inside metadata.

    Returns:
      `DefaultTrialConverter`.
    """
    # Cache ParameterConfigs.
    # Traverse through all parameter configs and merge the same-named ones.
    parameter_configs: Dict[str, pyvizier.ParameterConfig] = dict()
    for study_config in study_configs:
      all_parameter_configs = itertools.chain.from_iterable([
          top_level_config.traverse()
          for top_level_config in study_config.search_space.parameters
      ])
      for parameter_config in all_parameter_configs:
        name = parameter_config.name  # Alias
        existing_config = parameter_configs.get(name, None)
        if existing_config is None:
          parameter_configs[name] = parameter_config
        else:
          parameter_configs[name] = pyvizier.ParameterConfig.merge(
              existing_config, parameter_config)

    parameter_converters = []
    for pc in parameter_configs.values():
      parameter_converters.append(DefaultModelInputConverter(pc))

    # Append study id feature if configured to do so.
    if use_study_id_feature:
      # Collect study ids.
      study_ids = set()
      for sc in study_configs:
        metalearn_study_id = sc.metadata.get(STUDY_ID_FIELD, None)
        if metalearn_study_id is None:
          continue
        study_ids.add(metalearn_study_id)

      # Validate.
      if not study_ids:
        logging.error('use_study_id_feature was True, but none of the studies '
                      'had study id configured.')
        use_study_id_feature = False
      elif STUDY_ID_FIELD in parameter_configs:
        raise ValueError('Dataset name conflicts with a ParameterConfig '
                         'that already exists: {}'.format(
                             parameter_configs[STUDY_ID_FIELD]))

      # Create new parameter config.
      parameter_config = pyvizier.ParameterConfig.factory(
          STUDY_ID_FIELD, feasible_values=list(study_ids))
      parameter_configs[STUDY_ID_FIELD] = parameter_config
      logging.info('Created a new ParameterConfig %s', parameter_config)

      # Create converter.
      parameter_converters.append(
          DefaultModelInputConverter(
              parameter_config,
              lambda t: t.metadata.get(STUDY_ID_FIELD, None),
              converts_to_parameter=False))

    return cls(parameter_converters,
               [DefaultModelOutputConverter(m) for m in metric_information])

  @classmethod
  def from_study_config(cls, study_config: pyvizier.StudyConfig):
    return cls.from_study_configs([study_config],
                                  study_config.metric_information,
                                  use_study_id_feature=False)


class TrialToArrayConverter:
  """TrialToArrayConverter.

  Use a factory method (currently, there is one: `from_study_config`) instead
  of `__init__`.

  Unlike TrialtoNumpyDict converters, `to_features` and `to_labels`
  return a single array of floating numbers. CATEGORICAL and DISCRETE parameters
  are one-hot embedded.
  """
  _waiver = 'I am aware that this code may break at any point.'

  def __init__(self, impl: DefaultTrialConverter, _waiver: str = ''):
    """SHOULD NOT BE USED! Use factory classmethods e.g. from_study_config."""

    if _waiver != self._waiver:
      raise ValueError('Sign the waiver if you want to use init directly.')
    self._impl = impl

  def to_features(self, trials) -> np.ndarray:
    return dict_to_array(self._impl.to_features(trials))

  def to_labels(self, trials) -> np.ndarray:
    return dict_to_array(self._impl.to_labels(trials))

  def to_xy(self, trials) -> Tuple[np.ndarray, np.ndarray]:
    return self.to_features(trials), self.to_labels(trials)

  def to_parameters(self, arr: np.ndarray) -> Sequence[pyvizier.ParameterDict]:
    """Convert to nearest feasible parameter value. NaNs are preserved."""
    # TODO: Add a boolean flag to disable automatic clipping.
    arrformat = DictOf2DArrays(self._impl.to_features([]))
    return self._impl.to_parameters(arrformat.dict_like(arr))

  @classmethod
  def from_study_config(cls,
                        study_config: pyvizier.StudyConfig,
                        *,
                        scale: bool = True,
                        pad_oovs: bool = True,
                        max_discrete_indices: int = 0,
                        flip_sign_for_minimization_metrics: bool = True,
                        dtype=np.float64) -> 'TrialToArrayConverter':
    """From study config.

    Args:
      study_config:
      scale: If True, scales the parameters to [0, 1] range.
      pad_oovs: If True, add an extra dimension for out-of-vocabulary values for
        non-CONTINUOUS parameters.
      max_discrete_indices: For DISCRETE and INTEGER types that have more than
        this many feasible values will be continuified. When generating
        suggestions, values are rounded to the nearest feasible value.
      flip_sign_for_minimization_metrics: If True, flips the metric signs so
        that every metric maximizes.
      dtype: dtype

    Returns:
      TrialToArrayConverter
    """

    def create_input_converter(parameter):
      return DefaultModelInputConverter(
          parameter,
          scale=scale,
          max_discrete_indices=max_discrete_indices,
          onehot_embed=True,
          float_dtype=dtype,
          pad_oovs=pad_oovs)

    def create_output_converter(metric):
      return DefaultModelOutputConverter(
          metric,
          flip_sign_for_minimization_metrics=flip_sign_for_minimization_metrics,
          dtype=dtype)

    sc = study_config  # alias, to keep pylint quiet in the next line.
    converter = DefaultTrialConverter(
        [create_input_converter(p) for p in sc.search_space.parameters],
        [create_output_converter(m) for m in sc.metric_information])

    return cls(converter, cls._waiver)

  @property
  def output_specs(self) -> Sequence[NumpyArraySpec]:
    return [
        converter.output_spec for converter in self._impl.parameter_converters
    ]



"""Large-scale stress tests (multiple clients, mulithreading, etc.) for Vizier Service."""

import multiprocessing.pool
import time
from absl import logging

from vizier import benchmarks
from vizier.service import pyvizier
from vizier.service import vizier_client
from vizier.service.testing import local_service

from absl.testing import absltest
from absl.testing import parameterized


class PerformanceTest(parameterized.TestCase):
  service: local_service.LocalVizierTestService

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    cls.service = local_service.LocalVizierTestService()

  @parameterized.parameters(
      (1, 10, 2),
      (2, 10, 2),
      (10, 10, 2),
  )
  def test_multiple_clients_basic(self, num_simultaneous_clients,
                                  num_trials_per_client, dimension):

    def fn(client_id: int):
      func = benchmarks.bbob.Sphere
      experimenter = benchmarks.NumpyExperimenter(
          func, benchmarks.bbob.DefaultBBOBProblemStatement(dimension))
      problem_statement = experimenter.problem_statement()
      study_config = pyvizier.StudyConfig.from_problem(problem_statement)
      study_config.algorithm = pyvizier.Algorithm.RANDOM_SEARCH

      client = vizier_client.create_or_load_study(
          service_endpoint=self.service.endpoint,
          owner_id='my_username',
          study_id=self.id(),  # Use the testcase name.
          study_config=study_config,
          client_id=str(client_id))

      for _ in range(num_trials_per_client):
        suggestions = client.get_suggestions(suggestion_count=1)
        completed_trials = experimenter.evaluate(suggestions)
        for completed_trial in completed_trials:
          client.complete_trial(completed_trial.id,
                                completed_trial.final_measurement)

      return client

    client_ids = range(num_simultaneous_clients)
    pool = multiprocessing.pool.ThreadPool(num_simultaneous_clients)

    start = time.time()
    clients = pool.map(fn, client_ids)
    end = time.time()
    pool.close()

    self.assertEqual(
        max({t.id for t in clients[0].list_trials()}),
        num_simultaneous_clients * num_trials_per_client)

    logging.info(
        'For %d clients to evaluate %d trials each, it took %f seconds total.',
        num_simultaneous_clients, num_trials_per_client, end - start)


if __name__ == '__main__':
  absltest.main()



"""Tests for clients."""

from absl import flags
from absl import logging
import grpc
from vizier._src.pyvizier.client import client_abc_testing
from vizier.service import clients
from vizier.service import pyvizier as vz
from vizier.service.testing import local_service

from absl.testing import absltest

FLAGS = flags.FLAGS


class VizierClientTest(client_abc_testing.TestCase):
  _service: local_service.LocalVizierTestService
  _owner: str
  _channel: grpc.Channel

  @classmethod
  def setUpClass(cls):
    logging.info('Test setup started.')
    super().setUpClass()
    cls._service = local_service.LocalVizierTestService()
    clients.environment_variables.service_endpoint = cls._service.endpoint
    logging.info('Test setup finished.')

  def create_study(self, problem: vz.ProblemStatement,
                   study_id: str) -> clients.Study:
    config = vz.StudyConfig.from_problem(problem)
    config.algorithm = vz.Algorithm.NSGA2
    study = clients.Study.from_study_config(
        config, owner='owner', study_id=study_id)
    return study

  def test_e2e_tuning(self):
    self.assertPassesE2ETuning()

  @classmethod
  def tearDownClass(cls):
    cls._service._server.stop(None)
    super().tearDownClass()


if __name__ == '__main__':
  absltest.main()



"""Tests for core."""

import numpy as np
from vizier import pyvizier
from vizier._src.algorithms.designers import random
from vizier._src.algorithms.testing import test_runners
from vizier.pyvizier.converters import core
from vizier.testing import test_studies

from absl.testing import absltest
from absl.testing import parameterized

Trial = pyvizier.Trial


class TrialToArrayConverterTest(absltest.TestCase):
  """Test TrialToArrayConverter class."""

  def setUp(self):
    super().setUp()
    self._study_config = pyvizier.StudyConfig(
        search_space=test_studies.flat_space_with_all_types(),
        metric_information=[
            pyvizier.MetricInformation(
                'x1', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)
        ])

    self._designer = random.RandomDesigner(
        self._study_config.search_space, seed=0)
    self._trials = test_runners.run_with_random_metrics(
        self._designer, self._study_config, iters=1, batch_size=10)
    self.maxDiff = None  # pylint: disable=invalid-name

  def test_back_to_back_conversion(self):
    converter = core.TrialToArrayConverter.from_study_config(self._study_config)
    self.assertSequenceEqual([t.parameters for t in self._trials],
                             converter.to_parameters(
                                 converter.to_features(self._trials)))


class DictToArrayTest(absltest.TestCase):

  def test_2d(self):
    self.assertSequenceEqual(
        core.dict_to_array({
            1: np.random.random((10, 2)),
            2: np.random.random((10, 1))
        }).shape, (10, 3))

  def test_4d(self):
    self.assertSequenceEqual(
        core.dict_to_array({
            1: np.random.random((2, 1, 10, 2)),
            2: np.random.random((2, 1, 10, 1))
        }).shape, (2, 1, 10, 3))

  def test_dict_like(self):
    d = core.DictOf2DArrays({
        'p1': np.array([[1], [2], [3]]),
        'p2': np.array([[.0, .1], [.5, .5], [.1, .0]])
    })
    self.assertCountEqual(
        d.dict_like(np.array([[1, 1, .1], [3, 0, .1]])), {
            'p1': np.array([[1], [3]]),
            'p2': np.array([[1, .1], [0, .1]])
        })

  def test_dict_like_and_as_array_are_inverse(self):
    d = core.DictOf2DArrays({
        'p1': np.array([[1], [2], [3]]),
        'p2': np.array([[.0, .1], [.5, .5], [.1, .0]])
    })
    self.assertCountEqual(d, d.dict_like(d.asarray()))


rnd = np.random.random


class DictOf2DArraysTest(absltest.TestCase):

  def test_bad_shapes(self):
    with self.assertRaises(ValueError):
      core.DictOf2DArrays({'a': rnd([10, 2]), 'b': rnd([5, 2])})

  def test_bad_dim(self):
    with self.assertRaises(ValueError):
      core.DictOf2DArrays({'a': rnd([10, 2, 1])})

  def test_size(self):
    d = core.DictOf2DArrays({'a': rnd([10, 2]), 'b': rnd([10, 1])})
    self.assertEqual(d.size, 10)

  def test_addition(self):
    d1 = core.DictOf2DArrays({'a': rnd([10, 2]), 'b': rnd([10, 1])})
    d2 = core.DictOf2DArrays({'a': rnd([5, 2]), 'b': rnd([5, 1])})
    d = d1 + d2
    self.assertEqual(d.size, 15)
    self.assertLen(d, 2)
    self.assertSequenceEqual(d['a'].shape, (15, 2))
    self.assertSequenceEqual(d['b'].shape, (15, 1))

  def test_bad_addition(self):
    # Shape of b doesn't match.
    d1 = core.DictOf2DArrays({'a': rnd([10, 2]), 'b': rnd([10, 1])})
    d2 = core.DictOf2DArrays({'a': rnd([5, 2]), 'b': rnd([5, 3])})
    with self.assertRaises(ValueError):
      _ = d1 + d2


class DefaultTrialConverterFromStudyConfigsTest(absltest.TestCase):

  @property
  def _study_configs(self):
    study_configs = []
    space1 = pyvizier.SearchSpace()
    root = space1.select_root()
    root.add_float_param('double', -1., 1.)
    root.add_int_param('integer', -1, 1)
    root.add_categorical_param('categorical', ['a', 'b'])
    root.add_discrete_param('discrete', [-1., 1.])
    study_configs.append(
        pyvizier.StudyConfig(
            metadata=pyvizier.Metadata({core.STUDY_ID_FIELD: 'study1'}),
            search_space=space1))

    space2 = pyvizier.SearchSpace()
    root = space2.select_root()
    root.add_float_param('double', -1., 2.)
    root.add_int_param('integer', -2, 1)
    root.add_categorical_param('categorical', ['b', 'c'])
    root.add_discrete_param('discrete', [-1., 1., 2.])
    study_configs.append(
        pyvizier.StudyConfig(
            metadata=pyvizier.Metadata({core.STUDY_ID_FIELD: 'study2'}),
            search_space=space2))
    return study_configs

  def test_parameters(self):
    converter = core.DefaultTrialConverter.from_study_configs(
        self._study_configs, [], use_study_id_feature=True)
    self.assertCountEqual(
        ['study1', 'study2'],
        converter.parameter_configs[core.STUDY_ID_FIELD].feasible_values)
    self.assertEqual((-1., 2.), converter.parameter_configs['double'].bounds)
    self.assertEqual((-2, 1), converter.parameter_configs['integer'].bounds)
    self.assertCountEqual(
        ('a', 'b', 'c'),
        converter.parameter_configs['categorical'].feasible_values)
    self.assertCountEqual(
        (-1., 1., 2.), converter.parameter_configs['discrete'].feasible_values)

    expected = {
        'metalearn_study_id': (None, 1),
        'double': (None, 1),
        'integer': (None, 1),
        'categorical': (None, 1),
        'discrete': (None, 1),
    }
    self.assertDictEqual(converter.features_shape, expected)

    trial = pyvizier.Trial(
        parameters={
            'double': pyvizier.ParameterValue(3.),
            'integer': pyvizier.ParameterValue(-1),
            'categorical': pyvizier.ParameterValue('d')
        })
    expected = {
        'categorical': np.array([[3]]),
        'discrete': np.array([[3]]),
        'double': np.array([[3.]]),
        'integer': np.array([[1]]),
        'metalearn_study_id': np.array([[2]])
    }
    self.assertDictEqual(converter.to_features([trial]), expected)

    # The value for `categorical` was invalid, so we can't recover it.
    expected_parameters = pyvizier.ParameterDict({
        'double': pyvizier.ParameterValue(2.),
        'integer': pyvizier.ParameterValue(-1),
    })
    self.assertEqual(
        converter.to_trials(expected)[0].parameters, expected_parameters)

  def test_metrics(self):
    converter = core.DefaultTrialConverter.from_study_configs(
        [], [
            pyvizier.MetricInformation(
                name='metric1', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE)
        ],
        use_study_id_feature=False)

    expected = {'metric1': (None, 1)}
    self.assertDictEqual(converter.labels_shape, expected)


class DefaultModelOutputConverterTest(parameterized.TestCase):

  @property
  def _measurements(self):
    return [
        pyvizier.Measurement(metrics={
            'metric1': pyvizier.Metric(1.0),
            'metric2': pyvizier.Metric(1.1)
        }),
        pyvizier.Measurement(metrics={
            'metric1': pyvizier.Metric(2.0),
            'metric2': pyvizier.Metric(2.1)
        }),
        pyvizier.Measurement(metrics={
            'metric1': pyvizier.Metric(4.0),
            'metric2': pyvizier.Metric(4.1)
        }),
    ]

  def test_empty(self):
    converter = core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='metric1', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE),
        flip_sign_for_minimization_metrics=True)
    np.testing.assert_array_equal(
        converter.convert([]), np.zeros([0, 1], dtype=converter.dtype))

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_sign_flips(self, dtype):
    converter = core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='metric1', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE),
        flip_sign_for_minimization_metrics=True,
        dtype=dtype)
    actual = converter.convert(self._measurements)

    expected = -np.asarray([[1.], [2.], [4.]], dtype)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)
    self.assertEqual(
        converter.metric_information,
        pyvizier.MetricInformation(
            name='metric1', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE))

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_no_sign_flips(self, dtype):
    converter = core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='metric2', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE),
        flip_sign_for_minimization_metrics=False,
        dtype=dtype)
    actual = converter.convert(self._measurements)

    expected = np.asarray([[1.1], [2.1], [4.1]], dtype)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)
    self.assertEqual(
        converter.metric_information,
        pyvizier.MetricInformation(
            name='metric2', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))

  def test_raise_errors_for_missing_metrics(self):
    converter = core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='metric3', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE),
        flip_sign_for_minimization_metrics=False,
        raise_errors_for_missing_metrics=True)
    with self.assertRaises(KeyError):
      converter.convert(self._measurements)

  def test_do_not_raise_errors_for_missing_metrics(self):
    converter = core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='metric3', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE),
        flip_sign_for_minimization_metrics=False,
        raise_errors_for_missing_metrics=False)
    np.testing.assert_equal(
        converter.convert(self._measurements), np.asarray([[np.nan]] * 3))

  @parameterized.parameters([dict(flip_sign=True), dict(flip_sign=False)])
  def test_safe_maximize_metric(self, flip_sign):
    converter = core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='metric1',
            goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE,
            safety_threshold=3.),
        flip_sign_for_minimization_metrics=flip_sign)
    actual = converter.convert(self._measurements)
    np.testing.assert_equal(actual, [[-2.], [-1.], [1.]])

  @parameterized.parameters([dict(flip_sign=True), dict(flip_sign=False)])
  def test_safe_minimize_metric(self, flip_sign):
    converter = core.DefaultModelOutputConverter(
        pyvizier.MetricInformation(
            name='metric1',
            goal=pyvizier.ObjectiveMetricGoal.MINIMIZE,
            safety_threshold=3.),
        flip_sign_for_minimization_metrics=flip_sign)
    actual = converter.convert(self._measurements)
    expected = np.array([[-2.], [-1.], [1.]],
                        dtype=np.float32) * (-1 if flip_sign else 1)
    np.testing.assert_equal(actual, expected)


class DefaultModelInputConverterTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_double_into_double(self, dtype):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', bounds=(-3., 3.)),
        scale=False,
        onehot_embed=True,
        float_dtype=dtype)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(2)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(-3)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[1.0], [2.0], [-3.0], [np.NaN], [np.NaN]], dtype)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_double_into_double_log(self, dtype):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory(
            'x1', bounds=(1e-4, 1e2), scale_type=pyvizier.ScaleType.LOG),
        scale=True,
        float_dtype=dtype)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1e-4)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(1e2)}),
    ])
    expected = np.asarray([[0.0], [1.0]], dtype)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_double_into_double_log_inverse(self, dtype):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory(
            'x1', bounds=(1e-4, 1e2), scale_type=pyvizier.ScaleType.LOG),
        scale=True,
        float_dtype=dtype)

    scaled = np.asarray([[0.0], [1.0]], dtype)
    actual = converter.to_parameter_values(scaled)
    self.assertGreaterEqual(actual[0].value, 1e-4)
    self.assertLessEqual(actual[1].value, 1e2)

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_zero_range_linear_double(self, dtype):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory(
            'x1', bounds=(.9, .9), scale_type=pyvizier.ScaleType.LINEAR),
        scale=True,
        onehot_embed=True,
        float_dtype=dtype)
    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(.9)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(1.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[.0], [.0], [np.NaN], [np.NaN]], dtype)
    np.testing.assert_equal(expected, actual)

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_zero_range_log_double(self, dtype):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory(
            'x1',
            bounds=(np.exp(.9), np.exp(.9)),
            scale_type=pyvizier.ScaleType.LOG),
        scale=True,
        onehot_embed=True,
        float_dtype=dtype)
    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(np.exp(.9))}),
        Trial(parameters={'x1': pyvizier.ParameterValue(np.exp(1.2))}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[.0], [.0], [np.NaN], [np.NaN]], dtype=dtype)
    np.testing.assert_equal(expected, actual)

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_double_into_scaled_double(self, dtype):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory(
            'x1', bounds=(-3., 3.), scale_type=pyvizier.ScaleType.LINEAR),
        scale=True,
        onehot_embed=True,
        float_dtype=dtype)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(2)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(-3)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[4 / 6], [5 / 6], [0 / 6], [np.NaN], [np.NaN]],
                          dtype=dtype)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)

  def test_integer_discretes_into_discretes(self):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', feasible_values=(1, 2, 3)),
        max_discrete_indices=10)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(2)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(-3)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[0], [1], [3], [3], [3]], np.int32)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)

  @parameterized.parameters([
      dict(dtype=np.float32),
      dict(dtype=np.float64),
      dict(dtype='float32'),
      dict(dtype='float64')
  ])
  def test_integer_discretes_into_doubles(self, dtype):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', feasible_values=(1, 2, 3)),
        max_discrete_indices=1,
        float_dtype=dtype)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(2)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(-3)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[1.0], [2.0], [-3.0], [np.NaN], [np.NaN]],
                          dtype=dtype)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)

  def test_integer_discretes_into_onehot(self):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', feasible_values=(1, 2, 3)),
        scale=True,
        onehot_embed=True,
        max_discrete_indices=10)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(2)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(-3)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([
        [1., 0., 0., 0.],
        [0., 1., 0., 0.],
        [0., 0., 0., 1.],
        [0., 0., 0., 1.],
        [0., 0., 0., 1.],
    ],
                          dtype=np.float32)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)
    self.assertEqual(
        converter.output_spec,
        core.NumpyArraySpec(core.NumpyArraySpecType.ONEHOT_EMBEDDING,
                            np.float32, (0, 1), 4, 'x1', 1),
        msg=repr(converter.output_spec))

  def test_integer_discretes_into_onehot_empty(self):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', feasible_values=(1, 2, 3)),
        scale=True,
        onehot_embed=True,
        max_discrete_indices=10)

    actual = converter.convert([])
    expected = np.zeros([0, 4], dtype=np.float32)
    np.testing.assert_allclose(expected, actual)
    self.assertEqual(expected.dtype, actual.dtype)

  def test_discretes_into_discretes(self):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', feasible_values=(1., 2., 3.)),
        max_discrete_indices=10)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(2.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(-3.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[0], [1], [3], [3], [3]], dtype=np.float32)
    np.testing.assert_equal(expected, actual)

  def test_discretes_into_doubles(self):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', feasible_values=(1., 2., 3.)),
        max_discrete_indices=1)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(2.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(-3.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[1.0], [2.0], [-3.0], [np.NaN], [np.NaN]],
                          dtype=np.float32)
    np.testing.assert_equal(expected, actual)

  def test_discretes_into_scaled_doubles(self):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', feasible_values=(1., 2., 3.)),
        max_discrete_indices=1,
        scale=True)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue(1.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(2.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue(-3.)}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[0.], [.5], [-2.], [np.NaN], [np.NaN]],
                          dtype=np.float32)
    np.testing.assert_equal(expected, actual)

  def test_categorical(self):
    converter = core.DefaultModelInputConverter(
        pyvizier.ParameterConfig.factory('x1', feasible_values=('1', '2', '3')),
        max_discrete_indices=1)

    actual = converter.convert([
        Trial(parameters={'x1': pyvizier.ParameterValue('1')}),
        Trial(parameters={'x1': pyvizier.ParameterValue('2')}),
        Trial(parameters={'x1': pyvizier.ParameterValue('-3')}),
        Trial(parameters={'x1': pyvizier.ParameterValue('a')}),
        Trial()
    ])
    expected = np.asarray([[0], [1], [3], [3], [3]], dtype=np.float32)
    np.testing.assert_equal(expected, actual)


if __name__ == '__main__':
  absltest.main()



"""These are the externally-useful symbols for Pythia."""
# pylint: disable=unused-import

# The API for a Pythia Policy -- i.e. the algorithm that Pythia serves.
from vizier._src.pythia.local_policy_supporters import LocalPolicyRunner
from vizier._src.pythia.policy import EarlyStopDecision
from vizier._src.pythia.policy import EarlyStopRequest
from vizier._src.pythia.policy import Policy
from vizier._src.pythia.policy import SuggestDecision
from vizier._src.pythia.policy import SuggestDecisions
from vizier._src.pythia.policy import SuggestRequest
from vizier._src.pythia.policy_supporter import MetadataDelta
from vizier._src.pythia.policy_supporter import PolicySupporter
from vizier._src.pythia.pythia_errors import CachedPolicyIsStaleError
from vizier._src.pythia.pythia_errors import CancelComputeError
from vizier._src.pythia.pythia_errors import CancelledByVizierError
from vizier._src.pythia.pythia_errors import InactivateStudyError
from vizier._src.pythia.pythia_errors import LoadTooLargeError
from vizier._src.pythia.pythia_errors import PythiaError
from vizier._src.pythia.pythia_errors import PythiaProtocolError
from vizier._src.pythia.pythia_errors import TemporaryPythiaError
from vizier._src.pythia.pythia_errors import VizierDatabaseError
